/***************************************************************************
                          buttonwidget.cpp - eboxy button widget
                             -------------------
    begin                : Sat Aug 3 2002
    copyright            : (C) 2002 by Paul Eggleton
    email                : bluelightning@bluelightning.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <iostream>
#include <cstring>
#include <cstdlib>
#include "SDL_gui.h"

#include "utils.h"
#include "gui_extensions.h"
#include "actionmanager.h"
#include "errors.h"
#include "main.h"

#include "buttonwidget.h"
#include "page.h"

using namespace std;

extern ebPage *currentPage;
extern ActionManager action_manager;
extern const int DEFAULT_FONTSIZE;
extern const SDL_Color DEFAULT_FONT_COLOR;
extern const SDL_Color DEFAULT_SELECTED_FONT_COLOR;
extern const char *TYPESTR_BUTTON;

ButtonWidget::ButtonWidget(const char *name, int x, int y, int w, int h)
: ebWidget(name, x, y) {
  this->w = w;
  this->h = h;
  // Set defaults
  tabstop = true;
  font = NULL;
  action = NULL;
  caption = NULL;
  captionlabel = NULL;
  captionx = 0;
  captiony = 0;
  captiondropx = 2;
  captiondropy = 2;
  normalimage = NULL;
  selectedimage = NULL;
  pressedimage = NULL;
  disabledimage = NULL;
  fontcolor = colordup(&DEFAULT_FONT_COLOR);
  fontsize = DEFAULT_FONTSIZE;
  fontfocuscolor = NULL;
  fontdisabledcolor = NULL;
  fontpressedcolor = NULL;
  enabled = true;
  this->type = TYPESTR_BUTTON;

  AddProperty("caption", ButtonWidget_GetCaption, ButtonWidget_SetCaption);
  AddProperty("font", ButtonWidget_GetFont, ButtonWidget_SetFont);
  AddProperty("fontsize", ButtonWidget_GetFontSize, ButtonWidget_SetFontSize);
  AddProperty("fontcolor", ButtonWidget_GetFontColor, ButtonWidget_SetFontColor);
  AddProperty("fontfocuscolor", ButtonWidget_GetFontFocusColor, ButtonWidget_SetFontFocusColor);
  AddProperty("fontpressedcolor", ButtonWidget_GetFontPressedColor, ButtonWidget_SetFontPressedColor);
  AddProperty("fontdisabledcolor", ButtonWidget_GetFontDisabledColor, ButtonWidget_SetFontDisabledColor);
  AddProperty("image", ButtonWidget_GetImage, ButtonWidget_SetImage);
  AddProperty("selectedimage", ButtonWidget_GetSelectedImage, ButtonWidget_SetSelectedImage);
  AddProperty("pressedimage", ButtonWidget_GetPressedImage, ButtonWidget_SetPressedImage);
  AddProperty("disabledimage", ButtonWidget_GetDisabledImage, ButtonWidget_SetDisabledImage);
  AddProperty("captionx", ButtonWidget_GetCaptionX, ButtonWidget_SetCaptionX);
  AddProperty("captiony", ButtonWidget_GetCaptionY, ButtonWidget_SetCaptionY);
  AddProperty("captiondropx", ButtonWidget_GetCaptionDropX, ButtonWidget_SetCaptionDropX);
  AddProperty("captiondropy", ButtonWidget_GetCaptionDropY, ButtonWidget_SetCaptionDropY);
  AddProperty("enabled", ButtonWidget_GetEnabled, ButtonWidget_SetEnabled);
  AddProperty("action", ButtonWidget_GetAction, ButtonWidget_SetAction);

  AddMethod("click", 0, ButtonWidget_Click);
  
  AddEvent("OnClick");
  AddEvent("OnGotFocus");
  AddEvent("OnLostFocus");
  defaultevent = 0;
}

ButtonWidget::~ButtonWidget(void) {
  if(this->caption)
    free(this->caption);
  if(this->fontcolor)
    delete this->fontcolor;
  if(this->fontfocuscolor)
    delete this->fontfocuscolor;
  if(this->fontdisabledcolor)
    delete this->fontdisabledcolor;
  if(this->fontpressedcolor)
    delete this->fontpressedcolor;
}

GUI_Widget *ButtonWidget::Instantiate(void) {
  if(w<0) w=0;
  if(h<0) h=0;
  
  GUI_ExtButton *button = new GUI_ExtButton(name, x, y, w, h);

  SetupCaptionLabel(button);

  if(normalimage) {
    button->SetNormalImage(normalimage->surface);
    button->SetHighlightImage(normalimage->surface);  // this is not a typo - this helps with mouse support
  }
  if(pressedimage)
    button->SetPressedImage(pressedimage->surface);
  if(disabledimage)
    button->SetDisabledImage(disabledimage->surface);

  if(!enabled)
    button->SetFlags(WIDGET_DISABLED);
  else
    button->ClearFlags(WIDGET_DISABLED);

  GUI_EventHandler<ButtonWidget> *cb = new GUI_EventHandler<ButtonWidget>(this, &ButtonWidget::OnClick);
  button->SetClick(cb);
	cb->DecRef();

  cb = new GUI_EventHandler<ButtonWidget>(this, &ButtonWidget::OnUpDown);
  button->SetUpDownCallback(cb);
	cb->DecRef();

  if(action) {
    int rc = action_manager.RegisterLocalAction(action, this);
    if(rc == EBERR_ACTION_INVALID) {
      cerr << "Warning: action \"" << action << "\" is not a valid user action" << endl;
    }
    else if(rc == EBERR_ACTION_ASSIGNED) {
      cerr << "Warning: action \"" << action << "\" is assigned to more than one widget in this context" << endl;
    }
  }

  realwidget = button;
  return button;
}

void ButtonWidget::Uninstantiate(void) {
  if(realwidget) {
    if(captionlabel) {
      captionlabel->DecRef();
      captionlabel = NULL;
    }
    focused = false;

    if(action)
      action_manager.UnregisterLocalAction(action, this);
  }
  ebWidget::Uninstantiate();
}

void ButtonWidget::SetNormalImage(ImageEntry *image) {
  if(image) {
    this->normalimage = image;
    if(realwidget)
      ((GUI_ExtButton *)realwidget)->SetNormalImage(normalimage->surface);
    if(this->w=-1)
      SetWidth(-1);
    if(this->h=-1)
      SetHeight(-1);
  }
}

void ButtonWidget::SetSelectedImage(ImageEntry *image) {
  if(image) {
    this->selectedimage = image;
  }
}

void ButtonWidget::SetPressedImage(ImageEntry *image) {
  if(image) {
    this->pressedimage = image;
    if(realwidget)
      ((GUI_ExtButton *)realwidget)->SetPressedImage(pressedimage->surface);
  }
}

void ButtonWidget::SetDisabledImage(ImageEntry *image) {
  if(image) {
    this->disabledimage = image;
    if(realwidget)
      ((GUI_ExtButton *)realwidget)->SetDisabledImage(disabledimage->surface);
  }
}

void ButtonWidget::SetCaption(const char *caption) {
  if(this->caption)
    free(this->caption);
  this->caption = strdup(caption);
  if(realwidget) {
    if(!captionlabel)
      SetupCaptionLabel(realwidget);

    if(captionlabel) {
      captionlabel->SetText(caption);
      realwidget->MarkChanged();
    }
  }
}

void ButtonWidget::SetCaptionX(int value) {
  this->captionx = value;
  if(realwidget && captionlabel) {
    SDL_Rect r = captionlabel->GetArea();
    captionlabel->SetPosition(value, r.y);
  }
}

void ButtonWidget::SetCaptionY(int value) {
  this->captiony = value;
  if(realwidget && captionlabel) {
    SDL_Rect r = captionlabel->GetArea();
    captionlabel->SetPosition(r.x, value);
  }
}

void ButtonWidget::SetCaptionDropX(int value) {
  this->captiondropx = value;
  if(realwidget)
    ((GUI_ExtButton *)realwidget)->SetCaptionDropX(value);
}

void ButtonWidget::SetCaptionDropY(int value) {
  this->captiondropy = value;
  if(realwidget)
    ((GUI_ExtButton *)realwidget)->SetCaptionDropY(value);
}

void ButtonWidget::SetFont(FontEntry *font) {
  this->font = font;
  if(realwidget) {
    if(!captionlabel)
      SetupCaptionLabel(realwidget);

    if(captionlabel) {
      captionlabel->SetFont(this->font->font);
      captionlabel->SetText(this->caption); // fix for bug in SDL_GUI
      realwidget->MarkChanged();
    }
  }
}

void ButtonWidget::SetFontColor(const SDL_Color *c) {
  if(this->fontcolor)
    delete this->fontcolor;
  if(c)
    this->fontcolor = colordup(c);
  else
    this->fontcolor = NULL;
  if(realwidget && captionlabel)
    UpdateLabelColor(realwidget);
}

void ButtonWidget::SetFontFocusColor(const SDL_Color *c) {
  if(this->fontfocuscolor)
    delete this->fontfocuscolor;
  if(c)
    this->fontfocuscolor = colordup(c);
  else
    this->fontfocuscolor = NULL;
  if(realwidget && captionlabel)
    UpdateLabelColor(realwidget);
}

void ButtonWidget::SetFontDisabledColor(const SDL_Color *c) {
  if(this->fontdisabledcolor)
    delete this->fontdisabledcolor;
  if(c)
    this->fontdisabledcolor = colordup(c);
  else
    this->fontdisabledcolor = NULL;
  if(realwidget && captionlabel)
    UpdateLabelColor(realwidget);
}

void ButtonWidget::SetFontPressedColor(const SDL_Color *c) {
  if(this->fontpressedcolor)
    delete this->fontpressedcolor;
  if(c)
    this->fontpressedcolor = colordup(c);
  else
    this->fontpressedcolor = NULL;
  if(realwidget && captionlabel)
    UpdateLabelColor(realwidget);
}

void ButtonWidget::SetAction(const char *actionname) {
  this->action = GetUserAction(actionname);
}

void ButtonWidget::SetEnabled(bool enabled) {
  this->enabled = enabled;
  if(realwidget) {
    if(!enabled) {
      ((GUI_ExtButton *)realwidget)->SetFlags(WIDGET_DISABLED);
      if(focused && currentPage != NULL) {
        // If the widget is disabled and has the focus, we want to give the focus to some other widget
        currentPage->UpdateFocus();
      }
    }
    else
      ((GUI_ExtButton *)realwidget)->ClearFlags(WIDGET_DISABLED);
    if(captionlabel)
      UpdateLabelColor(realwidget);
  }
}

void ButtonWidget::SetWidth(int value) {
  if(value == -1 && normalimage != NULL)
    value = normalimage->surface->GetWidth();
  ebWidget::SetWidth(value);
} 

void ButtonWidget::SetHeight(int value) {
  if(value == -1 && normalimage != NULL)
    value = normalimage->surface->GetHeight();
  ebWidget::SetHeight(value);
}

void ButtonWidget::OnClick(GUI_Object *sender) {
  this->FireEvent("OnClick");
}

void ButtonWidget::OnUpDown(GUI_Object *sender) {
  if(captionlabel)
    UpdateLabelColor(realwidget);
}

void ButtonWidget::SimulateClick(void) {
  if(realwidget)
    ((GUI_ExtButton *)realwidget)->SimulateClick();
}

void ButtonWidget::GotFocus(void) {
  if(!focused) {
    this->FireEvent("OnGotFocus");
    ebWidget::GotFocus();
    if(selectedimage) {
      ((GUI_Button *)realwidget)->SetNormalImage(selectedimage->surface);
      ((GUI_Button *)realwidget)->SetHighlightImage(selectedimage->surface);
    }
    else if(normalimage) {
      ((GUI_Button *)realwidget)->SetNormalImage(normalimage->surface);
      ((GUI_Button *)realwidget)->SetHighlightImage(normalimage->surface);
    }
    else {
      ((GUI_Button *)realwidget)->SetNormalImage(NULL);
      ((GUI_Button *)realwidget)->SetHighlightImage(NULL);
    }
    if(captionlabel)
      UpdateLabelColor(realwidget);
  }
}

void ButtonWidget::LostFocus(void) {
  if(focused) {
    this->FireEvent("OnLostFocus");
    ebWidget::LostFocus();
    if(normalimage) {
      ((GUI_Button *)realwidget)->SetNormalImage(normalimage->surface);
      ((GUI_Button *)realwidget)->SetHighlightImage(normalimage->surface);
    }
    else {
      ((GUI_Button *)realwidget)->SetNormalImage(NULL);
      ((GUI_Button *)realwidget)->SetHighlightImage(NULL);
    }
    if(captionlabel)
      UpdateLabelColor(realwidget);
  }
}

void ButtonWidget::Execute(void) {
  SimulateClick();
}

void ButtonWidget::UpdateLabelColor(GUI_Widget *widget) {
  SDL_Color *c;

  if(enabled) {
    if(((GUI_ExtButton *)widget)->IsDown() && fontpressedcolor)
      c = fontpressedcolor;
    else if((widget->GetFlags() & WIDGET_INSIDE) != 0 && fontfocuscolor)
      c = fontfocuscolor;
    else
      c = fontcolor;
  }
  else {
    if(fontdisabledcolor)
      c = fontdisabledcolor;
    else
      c = fontcolor;
  }
  ((GUI_Label *)captionlabel)->SetTextColor(c->r, c->g, c->b);
  ((GUI_Label *)captionlabel)->SetText(caption); // fix for bug in SDL_GUI labelwidget
  widget->MarkChanged();
}

void ButtonWidget::SetupCaptionLabel(GUI_Widget *rwidget) {
  if (font != NULL && caption != NULL) {
    GUI_ExtButton *button = (GUI_ExtButton *)rwidget;

    // Using GUI_ExtLabel instead of GUI_Label solely because GUI_Label causes a crash if you
    // set the caption to an empty string (!)
		GUI_Label *label = new GUI_ExtLabel("lButtonLabel", captionx, captiony, w, h, font->font, caption);
		button->SetCaption(label);
		button->SetCaptionDropX(captiondropx);
		button->SetCaptionDropY(captiondropy);
    captionlabel = label;  // we don't decref() on the label here because we keep a reference
    UpdateLabelColor(button);
	}
}

/* Property get/set functions */

// -- caption --

const char *ButtonWidget_GetCaption(ebCustomObject *sender) {
  if( ((ButtonWidget *)sender)->caption )
    return ((ButtonWidget *)sender)->caption;
  else
    return "";  
}

void ButtonWidget_SetCaption(ebCustomObject *sender, const char *value) {
  ((ButtonWidget *)sender)->SetCaption(value);
}

// -- font --

const char *ButtonWidget_GetFont(ebCustomObject *sender) {
  return ((ButtonWidget *)sender)->font->filename;
}

void ButtonWidget_SetFont(ebCustomObject *sender, const char *value) {
  ButtonWidget *button = ((ButtonWidget *)sender);

  FontEntry *font = loadFont((char *)value, button->fontsize);
  if(font)
    button->SetFont(font);
}

// -- fontsize --

const char *ButtonWidget_GetFontSize(ebCustomObject *sender) {
  return inttostr(((ButtonWidget *)sender)->fontsize);
}

void ButtonWidget_SetFontSize(ebCustomObject *sender, const char *value) {
  ButtonWidget *button = ((ButtonWidget *)sender);
  int size = atoi(value);

  button->fontsize = size;
  if(button->font)
    ButtonWidget_SetFont(button, button->font->filename);
}

// -- fontcolor --

const char *ButtonWidget_GetFontColor(ebCustomObject *sender) {
  return colortostr(*(( (ButtonWidget *)sender)->fontcolor));
}

void ButtonWidget_SetFontColor(ebCustomObject *sender, const char *value) {
  SDL_Color c = strtocolor((char *)value);
  ((ButtonWidget *)sender)->SetFontColor(&c);
}

// -- fontfocuscolor --

const char *ButtonWidget_GetFontFocusColor(ebCustomObject *sender) {
  if( ((ButtonWidget *)sender)->fontfocuscolor)
    return colortostr(*( ((ButtonWidget *)sender)->fontfocuscolor));
  else
    return "";
}

void ButtonWidget_SetFontFocusColor(ebCustomObject *sender, const char *value) {
  if(strlen(value)) {
    SDL_Color c = strtocolor((char *)value);
    ((ButtonWidget *)sender)->SetFontFocusColor(&c);
  }
  else
    ((ButtonWidget *)sender)->SetFontFocusColor(NULL);
}

// -- fontpressedcolor --

const char *ButtonWidget_GetFontPressedColor(ebCustomObject *sender) {
  if( ((ButtonWidget *)sender)->fontpressedcolor )
    return colortostr(*( ((ButtonWidget *)sender)->fontpressedcolor ));
  else
    return "";
}

void ButtonWidget_SetFontPressedColor(ebCustomObject *sender, const char *value) {
  if(strlen(value)) {
    SDL_Color c = strtocolor((char *)value);
    ((ButtonWidget *)sender)->SetFontPressedColor(&c);
  }
  else
    ((ButtonWidget *)sender)->SetFontPressedColor(NULL);
}

// -- fontdisabledcolor --

const char *ButtonWidget_GetFontDisabledColor(ebCustomObject *sender) {
  if( ((ButtonWidget *)sender)->fontdisabledcolor )
    return colortostr(*( ((ButtonWidget *)sender)->fontdisabledcolor ));
  else
    return "";
}

void ButtonWidget_SetFontDisabledColor(ebCustomObject *sender, const char *value) {
  if(strlen(value)) {
    SDL_Color c = strtocolor((char *)value);
    ((ButtonWidget *)sender)->SetFontDisabledColor(&c);
  }
  else
    ((ButtonWidget *)sender)->SetFontDisabledColor(NULL);
}

// -- image --

const char *ButtonWidget_GetImage(ebCustomObject *sender) {
  if( ((ButtonWidget *)sender)->normalimage )
    return ((ButtonWidget *)sender)->normalimage->filename;
  else
    return "";
}

void ButtonWidget_SetImage(ebCustomObject *sender, const char *value) {
  ((ButtonWidget *)sender)->SetNormalImage(loadImage(value));
}

// -- selectedimage --

const char *ButtonWidget_GetSelectedImage(ebCustomObject *sender) {
  if( ((ButtonWidget *)sender)->selectedimage )
    return ((ButtonWidget *)sender)->selectedimage->filename;
  else
    return "";
}

void ButtonWidget_SetSelectedImage(ebCustomObject *sender, const char *value) {
  ((ButtonWidget *)sender)->SetSelectedImage(loadImage(value));
}

// -- pressedimage --

const char *ButtonWidget_GetPressedImage(ebCustomObject *sender) {
  if( ((ButtonWidget *)sender)->pressedimage )
    return ((ButtonWidget *)sender)->pressedimage->filename;
  else
    return "";
}

void ButtonWidget_SetPressedImage(ebCustomObject *sender, const char *value) {
  ((ButtonWidget *)sender)->SetPressedImage(loadImage(value));
}

// -- disabledimage --

const char *ButtonWidget_GetDisabledImage(ebCustomObject *sender) {
  if( ((ButtonWidget *)sender)->disabledimage )
    return ((ButtonWidget *)sender)->disabledimage->filename;
  else
    return "";
}

void ButtonWidget_SetDisabledImage(ebCustomObject *sender, const char *value) {
  ((ButtonWidget *)sender)->SetDisabledImage(loadImage(value));
}

// -- captionx --

const char *ButtonWidget_GetCaptionX(ebCustomObject *sender) {
  return inttostr( ((ButtonWidget *)sender)->captionx );
}

void ButtonWidget_SetCaptionX(ebCustomObject *sender, const char *value) {
  ((ButtonWidget *)sender)->SetCaptionX(atoi(value));
}

// -- captiony --

const char *ButtonWidget_GetCaptionY(ebCustomObject *sender) {
  return inttostr( ((ButtonWidget *)sender)->captiony );
}

void ButtonWidget_SetCaptionY(ebCustomObject *sender, const char *value) {
  ((ButtonWidget *)sender)->SetCaptionY(atoi(value));
}

// -- captiondropx --

const char *ButtonWidget_GetCaptionDropX(ebCustomObject *sender) {
  return inttostr( ((ButtonWidget *)sender)->captiondropx );
}

void ButtonWidget_SetCaptionDropX(ebCustomObject *sender, const char *value) {
  ((ButtonWidget *)sender)->SetCaptionDropX(atoi(value));
}

// -- captiondropy --

const char *ButtonWidget_GetCaptionDropY(ebCustomObject *sender) {
  return inttostr( ((ButtonWidget *)sender)->captiondropy );
}

void ButtonWidget_SetCaptionDropY(ebCustomObject *sender, const char *value) {
  ((ButtonWidget *)sender)->SetCaptionDropY(atoi(value));
}

// -- enabled --

const char *ButtonWidget_GetEnabled(ebCustomObject *sender) {
  return(booltostr( ((ButtonWidget *)sender)->enabled ));
}

void ButtonWidget_SetEnabled(ebCustomObject *sender, const char *value) {
  ((ButtonWidget *)sender)->SetEnabled(strtobool(value));
}

// -- action --

const char *ButtonWidget_GetAction(ebCustomObject *sender) {
  if( ((ButtonWidget *)sender)->action )
    return( ((ButtonWidget *)sender)->action );
  else
    return "";
}

void ButtonWidget_SetAction(ebCustomObject *sender, const char *value) {
  ((ButtonWidget *)sender)->SetAction(value);
}

/* Method functions */

const char *ButtonWidget_Click(ebCustomObject *sender, const char *method, int numargs, const char *args[]) {
  ((ButtonWidget *)sender)->SimulateClick();
  return "";
}
