/***************************************************************************
                          conf.cpp  -  configuration functions
                             -------------------
    begin                : Sat Nov 30 2002
    copyright            : (C) 2002 by Paul Eggleton
    email                : bluelightning@bluelightning.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <vector>
#include <iostream>
#include <fstream>

#include <ctype.h>
#include <errno.h>
#include <sys/stat.h>

#include "SDL.h"

#include "utils.h"
#include "conf.h"

#define CONF_DIR       "~/.eboxy/"
#define KEY_CONF_FILE  "~/.eboxy/keys"

using namespace std;

extern vector<keybinding> keybindings;

// The set of default key bindings (used to write first-time key bindings file)
// FIXME: shouldn't we be using our own MODKEY_ constants here instead of KMOD_?
keybinding default_keys[] = {
  {SDLK_ESCAPE,    KMOD_NONE, "quit"},
  {SDLK_RETURN,    KMOD_NONE, "accept"},
  {SDLK_LEFT,      KMOD_NONE, "prev"},
  {SDLK_RIGHT,     KMOD_NONE, "next"},
  {SDLK_UP,        KMOD_NONE, "item_prev"},
  {SDLK_DOWN,      KMOD_NONE, "item_next"},
  {SDLK_PAGEUP,    KMOD_NONE, "item_prevpage"},
  {SDLK_PAGEDOWN,  KMOD_NONE, "item_nextpage"},
  {SDLK_HOME,      KMOD_NONE, "item_first"},
  {SDLK_END,       KMOD_NONE, "item_last"},
  {SDLK_BACKSPACE, KMOD_NONE, NULL} // end of list
};

int readKeyConf(void) {
  char *conffile = expandPath(KEY_CONF_FILE);
  char buf[256];
  int bufptr = 0;
  keybinding bind;
  char ch;
  int modifier;
  char *keyname;

  char *confdir = NULL;

  switch(statfile(conffile)) {
    case ENOTDIR:
      // Fall through to code below
    case ENOENT:
      useDefaultBindings();
      // Try to create directory
      confdir = expandPath(CONF_DIR);
      if(mkdir(confdir, 0777) == -1 && errno != EEXIST) {
        cerr << "Failed to create directory " << confdir;
        perror("");
        free(confdir);
        free(conffile);
        return 0;
      }
      free(confdir);

      // Create the config file
      cout << "Creating config file " << conffile << endl;
      writeKeyConf(conffile);
      free(conffile);
      return 0;
      break;
    case EACCES:
      cerr << "Access denied to file " << conffile << endl;
      return 1;
      break;
    case 0:
      // File exists
      break;
    default:
      return 3;
  }

  ifstream conf_in(conffile, ios::in | ios::binary);

  if(!conf_in) {
    cerr << "Cannot open key config file " << conffile << endl;
    free(conffile);
    return 1;
  }

  // Now parse the config file

  enum {MODE_BEGIN, MODE_COMMENT, MODE_KEY, MODE_SEPARATOR, MODE_ACTION, MODE_EOL};
  int mode = MODE_BEGIN;
  int linecounter = 1;

  while(!conf_in.eof()) {
    // Get next character
    conf_in.get(ch);

    if(bufptr==256) {
      cerr << "Error in key config file " << conffile << " - line " << linecounter << " is too long" << endl;
      conf_in.close();
      free(conffile);
      return 2;
    }

    if(ch==0)
      ch = '\n';  // I know it's cheating, but who cares :)

    switch(mode) {
      case MODE_BEGIN:
        bufptr = 0;
        if(ch == '#')
          mode = MODE_COMMENT;
        else if(!isspace(ch)) {
          buf[bufptr++] = ch;
          mode = MODE_KEY;
        }
        break;
      case MODE_COMMENT:
        // In the middle of a comment, ignore anything but newlines
        if(ch == '\n')
          mode = MODE_BEGIN;
        break;
      case MODE_KEY:
        if(ch == '\n' || ch == '#') {
          cerr << "Error in key config file " << conffile << " - missing action at line " << linecounter << endl;
          conf_in.close();
          free(conffile);
          return 2;
        }
        else if(ch == ' ' || ch == '\t') {
          // End of key name
          buf[bufptr] = '\0';
          // Check if key is valid

          keyname = NULL;
          // FIXME: Lock keys are ignored right now (difficult to handle)
          //modifier = MODKEY_NONE;
          modifier = MODKEY_ANYLOCK;
          if(rindex(buf, '+')) {
            // Get modifiers
            char *keyspec = strdup(buf);
            char *p = keyspec;
            char *p2 = NULL;
            bool modfound = false;

            while(true) {
              p2 = strsep(&p, "+");
              if(!p2)
                break;

              modfound = false;
              int j=0;
              while(modkeynamelist[j].modstr != NULL) {
                if(!strcasecmp(p2, modkeynamelist[j].modstr)) {
                  modifier |= modkeynamelist[j].mod;
                  modfound = true;
                  break;
                }
                j++;
              }

              if(!modfound) {
                // If this is the last item then this is probably the key specifier itself
                keyname = strdup(p2);
                if(!strsep(&p, "+")) {
                  break;
                }
                else {
                  // Nope, error
                  free(keyspec);
                  free(keyname);
                  cerr << "Error in key config file " << conffile << " - \"" << buf << "\" not a valid key specifier (invalid modifier) at line " << linecounter << endl;
                  conf_in.close();
                  free(conffile);
                  return 2;
                }
              }
            }

            free(keyspec);
            if(!keyname) {
              cerr << "Error in key config file " << conffile << " - \"" << buf << "\" not a valid key specifier (missing key name) at line " << linecounter << endl;
              conf_in.close();
              free(conffile);
              return 2;
            }
          }
          else {
            keyname = strdup(buf);
          }

          bind.mod = (SDLMod)modifier;

          // Get the actual key specified
          int i=0;
          while(keynamelist[i].keystr != NULL) {
            if(!strcasecmp(keyname, keynamelist[i].keystr)) {
              bind.key = keynamelist[i].key;
              i = -1;
              break;
            }
            i++;
          }
          free(keyname);

          if(i!=-1) {
            cerr << "Error in key config file " << conffile << " - \"" << buf << "\" not a valid key specifier (invalid key name) at line " << linecounter << endl;
            conf_in.close();
            free(conffile);
            return 2;
          }
          // Into separator between key and action
          mode = MODE_SEPARATOR;
        }
        else
          buf[bufptr++] = ch;
        break;
      case MODE_SEPARATOR:
        bufptr = 0;
        if(ch == '\n' || ch == '#') {
          cerr << "Error in key config file " << conffile << " - missing action at line " << linecounter << endl;
          conf_in.close();
          free(conffile);
          return 2;
        }
        else if(!(ch == ' ' || ch == '\t')) {
          buf[bufptr++] = ch;
          mode = MODE_ACTION;
        }
        break;
      case MODE_ACTION:
        if(isspace(ch)) {
          // End of action
          buf[bufptr] = '\0';
          bind.action = strdup(buf);

          keybindings.push_back(bind);
          if(ch == '\n')
            mode = MODE_BEGIN;
          else
            mode = MODE_EOL;
        }
        else
          buf[bufptr++] = ch;
        break;
      case MODE_EOL:
        if(ch == '\n')
          mode = MODE_BEGIN;
        else if(ch == '#')
          mode = MODE_COMMENT;
        else if(!(ch == ' ' || ch == '\t')) {
          cerr << "Error in key config file " << conffile << " - unexpected characters at end of line " << linecounter << endl;
          conf_in.close();
          free(conffile);
          return 2;
        }
        break;
    }

    if(ch == '\n')
      linecounter++;
  }

  conf_in.close();

  free(conffile);
  return 0;
}

int writeKeyConf(const char *conffile) {
  ofstream conf_out(conffile);

  if(!conf_out) {
    cerr << "Cannot create key config file " << conffile << endl;
    return 1;
  }

  for(int i=0; i<keybindings.size(); i++) {
    for(int j=0; keynamelist[j].keystr != NULL; j++) {
      if(keynamelist[j].key == keybindings[i].key) {
        conf_out << keynamelist[j].keystr << "\t" << keybindings[i].action << endl;
        break;
      }
    }
  }

  conf_out.close();
  return 0;
}

void useDefaultBindings(void) {
  for(int i=0; default_keys[i].action != NULL; i++) {
    keybindings.push_back(default_keys[i]);
  }
}

char textInputNameToChar(const char *charname) {
  int i=0;
  while(textinputchars[i].charname != NULL) {
    if(!strcasecmp(charname, textinputchars[i].charname)) {
      return textinputchars[i].charchr;
    }
    i++;
  }
  return '\0';
}

const char *textInputCharToName(char c) {
  int i=0;
  while(textinputchars[i].charname != NULL) {
    if(textinputchars[i].charchr == c) {
      return textinputchars[i].charname;
    }
    i++;
  }
  return NULL;
}

const char *keyToName(SDLKey key) {
  for(int i=0; keynamelist[i].keystr != NULL; i++) {
    if(keynamelist[i].key == key)
      return keynamelist[i].keystr;
  }
  return "unknown";
}

bool checkmodkeys(SDLMod pressedmod, int expectedmod) {
  int modcode;
  bool expected;

  if(expectedmod==MODKEY_ANY)
    return true;  // don't care about modifiers
  else if(expectedmod==MODKEY_NONE && pressedmod==KMOD_NONE) {
    return true;
  }
  else {
    // Translate the SDL modifier bitmask into our own constant bitmask
    modcode = transmod(pressedmod);
    if(expectedmod==MODKEY_ANYLOCK) {
      if((modcode & (!MODKEY_LOCKMASK)) == 0)
        return true;
    }
    else if(modcode==expectedmod)
      return true;  // exact match
    else {
      for(int i=0; modkeynamelist[i].modstr != NULL; i++) {
        expected = ((expectedmod & modkeynamelist[i].mod) != 0);
        switch(modkeynamelist[i].mod) {
          case MODKEY_SHIFT:
            if(expected) {
              if((modcode & MODKEY_SHIFTMASK) == 0)
                return false;
              break;
            }
          case MODKEY_CTRL:
            if(expected) {
              if((modcode & MODKEY_CTRLMASK) == 0)
                return false;
              break;
            }
          case MODKEY_ALT:
            if(expected) {
              if((modcode & MODKEY_ALTMASK) == 0)
                return false;
              break;
            }
          case MODKEY_ANY:
          case MODKEY_ANYLOCK:
          case MODKEY_NONE:
            break;
          default:
            if(expected) {
              if((modcode & modkeynamelist[i].mod) == 0)
                return false;
            }
            else {
              if((modcode & modkeynamelist[i].mod) != 0) {
                if(!( ((expectedmod & MODKEY_SHIFT) != 0 && (modkeynamelist[i].mod & MODKEY_SHIFTMASK) != 0)
                    || ((expectedmod & MODKEY_CTRL) != 0 && (modkeynamelist[i].mod & MODKEY_CTRLMASK) != 0)
                    || ((expectedmod & MODKEY_ALT) != 0 && (modkeynamelist[i].mod & MODKEY_ALTMASK) != 0)))
                  return false;
              }
            }
            break;
        }
      }
      // Passed
      return true;
    }
  }
  return false;
}

int transmod(SDLMod sdlmod) {
  int mappedmod;

  mappedmod = 0;
  for(int i=0; modmap[i].modcode != MODKEY_NONE; i++) {
    if((sdlmod & modmap[i].sdlmod) != 0)
      mappedmod |= modmap[i].modcode;
  }

  return mappedmod;
}
