/***************************************************************************
                          customobject.cpp  -  eboxy custom object
                             -------------------
    begin                : Sun Sep 8 2002
    copyright            : (C) 2002 by Paul Eggleton
    email                : bluelightning@bluelightning.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <iostream>
#include <cstring>
#include <cstdlib>

#include "customobject.h"
#include "internalfunctionproperty.h"
#include "internalmethodhandler.h"
#include "errors.h"
#include "utils.h"

using namespace std;

ebCustomObject::ebCustomObject(const char *name)
: ebDynamicObject(name) {
  this->type = "CustomObject"; // If this is changed, need to change dynamicobject::ClearEventHandlers
  this->defaultproperty = NULL;

  AddProperty("name", CustomObject_GetName, NULL);
  AddProperty("type", CustomObject_GetType, NULL);
  AddProperty("propertycount", CustomObject_GetPropertyCount, NULL);
  AddProperty("methodcount", CustomObject_GetMethodCount, NULL);
  AddProperty("eventcount", CustomObject_GetEventCount, NULL);

  AddMethod("getproperty", 1, CustomObject_GetProperty);
  AddMethod("getmethod", 1, CustomObject_GetMethod);
  AddMethod("getevent", 1, CustomObject_GetEvent);
}

ebCustomObject::~ebCustomObject(void) {
  while(properties.size() > 0) {
    delete properties.back();
    properties.pop_back();
  }
}

int ebCustomObject::AddProperty(const char *name, const char *(*getter)(const char *), void (*setter)(const char *, const char *)) {
  if(HasProperty(name)) {
    cerr << "Object " << GetName() << " already has a property " << name << endl;
    return EBERR_MEMBER_EXISTS;
  }
  else if(HasMethod(name)) {
    cerr << "Object " << GetName() << " already has a method " << name << endl;
    return EBERR_MEMBER_EXISTS;
  }

  ebCustomProperty *prop = new ebCustomProperty(name, getter, setter);
  properties.push_back(prop);

  return EBERR_SUCCESS;
}

int ebCustomObject::AddProperty(const char *name, propertygetfn getter, propertysetfn setter, bool required) {
  if(HasProperty(name)) {
    cerr << "Object " << GetName() << " already has a property " << name << endl;
    return EBERR_MEMBER_EXISTS;
  }
  else if(HasMethod(name)) {
    cerr << "Object " << GetName() << " already has a method " << name << endl;
    return EBERR_MEMBER_EXISTS;
  }

  InternalFunctionProperty *prop = new InternalFunctionProperty(name, this, getter, setter, required);
  properties.push_back(prop);

  return EBERR_SUCCESS;
}

int ebCustomObject::AddMethod(const char *name, int numargs, methodfn methodfuncex) {
  // FIXME refactor out
  if(HasProperty(name)) {
    cerr << "Object " << GetName() << " already has a property " << name << endl;
    return EBERR_MEMBER_EXISTS;
  }
  else if(HasMethod(name)) {
    cerr << "Object " << GetName() << " already has a method " << name << endl;
    return EBERR_MEMBER_EXISTS;
  }

  InternalMethodHandler *mh = new InternalMethodHandler(name, this, numargs, methodfuncex);
  methodhandlers.push_back(mh);

  return EBERR_SUCCESS;
}

const char *ebCustomObject::GetPropertyValue(const char *property) {
  CustomProperty *prop = FindProperty(property);
  if(prop)
    return prop->GetValue();
  else {
    cerr << "Object " << GetName() << " does not have a property " << property << endl;
    return NULL;
  }
}

void ebCustomObject::SetPropertyValue(const char *property, const char *value) {
  CustomProperty *prop = FindProperty(property);
  if(prop)
    prop->SetValue(value);
  else
    cerr << "Object " << GetName() << " does not have a property " << property << endl;
}

CustomProperty *ebCustomObject::FindProperty(const char *name) {
  for(int i=0; i<properties.size(); i++) {
    if(!strcasecmp(properties[i]->GetName(), name))
      return properties[i];
  }
  return NULL;
}

CustomProperty *ebCustomObject::GetProperty(int index) {
  if(index < properties.size() && index >= 0)
    return properties[index];
  else
    return NULL;
}

int ebCustomObject::GetPropertyCount(void) {
  // of course, this DOES NOT count hacked in properties in child classes
  // (which is why they shouldn't be done that way)
  return properties.size();
}

const char *ebCustomObject::GetDefaultProperty(void) {
  if(defaultproperty)
    return defaultproperty->GetName();
  else
    return NULL;
}

bool ebCustomObject::HasProperty(const char *property) {
  return ((this->FindProperty(property)) != NULL);
}

int ebCustomObject::FireEvent(const char *event) {
  return ebDynamicObject::FireEvent(event);
}

int ebCustomObject::FireEventWait(const char *event) {
  return ebDynamicObject::FireEventWait(event);
}

/* Property get/set functions */

// -- name

const char *CustomObject_GetName(ebCustomObject *sender) {
  return sender->GetName();
}

// -- type

const char *CustomObject_GetType(ebCustomObject *sender) {
  return sender->GetType();
}

// -- propertycount

const char *CustomObject_GetPropertyCount(ebCustomObject *sender) {
  return inttostr(((ebCustomObject *)sender)->GetPropertyCount());
}

// -- methodcount

const char *CustomObject_GetMethodCount(ebCustomObject *sender) {
  return inttostr(((ebCustomObject *)sender)->GetMethodCount());
}

// -- eventcount

const char *CustomObject_GetEventCount(ebCustomObject *sender) {
  return inttostr(((ebCustomObject *)sender)->GetEventCount());
}

/* Method functions */

const char *CustomObject_GetProperty(ebCustomObject *sender, const char *method, int numargs, const char *args[]) {
  CustomProperty *prop = ((ebCustomObject *)sender)->GetProperty(atoi(args[0]));
  if(prop)
    return prop->GetName();
  else
    return "";
}

const char *CustomObject_GetMethod(ebCustomObject *sender, const char *method, int numargs, const char *args[]) {
  MethodHandler *mthd = ((ebCustomObject *)sender)->GetMethod(atoi(args[0]));
  if(mthd)
    return mthd->GetName();
  else
    return "";
}

const char *CustomObject_GetEvent(ebCustomObject *sender, const char *method, int numargs, const char *args[]) {
  const char *eventname = ((ebCustomObject *)sender)->GetEvent(atoi(args[0]));
  if(eventname)
    return eventname;
  else
    return "";
}
