/***************************************************************************
           dynamicobject.cpp  -  parent object class supporting events
                             -------------------
    begin                : Sun Sep 22 2002
    copyright            : (C) 2002 by Paul Eggleton
    email                : bluelightning@bluelightning.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <iostream>

#include "dynamicobject.h"
#include "functioneventhandler.h"
#include "scripteventhandler.h"
#include "pluginobject.h"
#include "errors.h"

using namespace std;

extern const char *TYPESTR_PLUGIN;

ebDynamicObject::ebDynamicObject(const char *name)
: ebObject(name) {
  this->type = "_DynamicObject";
  findable = true;
  container = NULL;
  defaultevent = -1;
}

ebDynamicObject::~ebDynamicObject(void) {
  ClearEventHandlers();
  ClearMethodHandlers();
  if(this->container)
    container->Notify(this);
}

EventHandler *ebDynamicObject::RegisterEventHandler(const char *event, ebObject *sink, int (*eventfunction)(const char *)) {
  EventHandler *handler = new FunctionEventHandler(event, this, sink, eventfunction);
  eventhandlers.push_back(handler);
  return handler;
}

EventHandler *ebDynamicObject::RegisterEventHandler(const char *event, Script *script) {
  EventHandler *handler = new ScriptEventHandler(event, this, NULL, script);
  eventhandlers.push_back(handler);
  return handler;
}

ExternalMethodHandler *ebDynamicObject::RegisterMethod(const char *method, ebObject *owner, int numargs, const char *(*methodfunc)(const char *, int, const char*[])) {
  ExternalMethodHandler *handler = new ExternalMethodHandler(method, this, numargs, methodfunc);
  methodhandlers.push_back(handler);
  return handler;
}

void ebDynamicObject::UnregisterEventHandler(EventHandler *handler) {
  for(int i=0;i<eventhandlers.size(); i++) {
    if(eventhandlers[i] == handler) {
      eventhandlers.erase(eventhandlers.begin() + i);
      delete handler;
      break;
    }
  }
}

int ebDynamicObject::UnregisterEventHandler(const char *event, ebObject *sink) {
  // Unregister all event handlers for the specified event registered by the specified sink
  int unregcount = 0;

  for(int i=0;i<eventhandlers.size();) {
    EventHandler *handler = eventhandlers[i];
    if(handler->GetSink() == sink && !strcasecmp(handler->GetName(), event)) {
      eventhandlers.erase(eventhandlers.begin() + i);
      delete handler;
      unregcount++;
    }
    else
      i++;
  }

  if(unregcount == 0)
    return EBERR_EVENT_NOT_HANDLED;
  else
    return EBERR_SUCCESS;
}

void ebDynamicObject::UnregisterMethod(MethodHandler *handler) {
  for(int i=0;i<methodhandlers.size(); i++) {
    if(methodhandlers[i] == handler) {
      methodhandlers.erase(methodhandlers.begin() + i);
      delete handler;
      break;
    }
  }
}

int ebDynamicObject::FireEvent(const char *event) {
  for(int i=0;i<eventhandlers.size(); i++) {
    if(!strcasecmp(eventhandlers[i]->GetName(), event)) {
      eventhandlers[i]->Fire(true);
    }
  }
  return 0;
}

int ebDynamicObject::FireEventWait(const char *event) {
  for(int i=0;i<eventhandlers.size(); i++) {
    if(!strcasecmp(eventhandlers[i]->GetName(), event))
      eventhandlers[i]->Fire(false);
  }
  return 0;
}

bool ebDynamicObject::EventHandled(const char *event) {
  for(int i=0;i<eventhandlers.size(); i++) {
    if(!strcasecmp(eventhandlers[i]->GetName(), event))
      return true;
  }
  return false;
}

const char *ebDynamicObject::MethodCall(const char *method, int numargs, const char *args[]) {
  for(int i=0;i<methodhandlers.size(); i++) {
    if(!strcasecmp(methodhandlers[i]->GetName(), method))
      return methodhandlers[i]->Call(numargs, args);
  }
  // Missing method
  cerr << "Object " << GetName() << " does not have a method " << method << endl;
  return "";
}

const char *ebDynamicObject::GetPropertyValue(const char *property) {
  if(!strcasecmp(property, "name"))
    return GetName();
  else if(!strcasecmp(property, "type"))
    return GetType();
  else {
    cerr << "Object " << GetName() << " does not have a property " << property << endl;
    return NULL;
  }
}

void ebDynamicObject::SetPropertyValue(const char *property, const char *value) {
  if(!strcasecmp(property, "name"))
    cerr << "Property \'name\' is read-only" << endl;
  else if(!strcasecmp(property, "type"))
    cerr << "Property \'type\' is read-only" << endl;
  else
    cerr << "Object " << GetName() << " does not have a property " << property << endl;
}

bool ebDynamicObject::HasProperty(const char *property) {
  if(!strcasecmp("name", property) || !strcasecmp("type", property))
    return true;
  else
    return false;
}

bool ebDynamicObject::HasMethod(const char *method) {
  for(int i=0;i<methodhandlers.size(); i++) {
    if(!strcasecmp(methodhandlers[i]->GetName(), method))
      return true;
  }
  return false;
}

MethodHandler *ebDynamicObject::GetMethod(int index) {
  if(index < methodhandlers.size() && index >= 0)
    return methodhandlers[index];
  else
    return NULL;
}

int ebDynamicObject::GetMethodCount(void) {
  // of course, this DOES NOT count hacked in properties in child classes
  // (which is why they shouldn't be done that way)
  return methodhandlers.size();
}

void ebDynamicObject::ClearEventHandlers(void) {
  // Remove all attached event handlers
  while(eventhandlers.size() > 0) {
    EventHandler *handler = eventhandlers.back();
    eventhandlers.pop_back();
    if(handler->GetSource() != NULL) {
      // Check for plugin-created objects
      if(!strcmp(handler->GetSource()->GetType(), "CustomObject")) {
        if(!strcmp(((ebDynamicObject *)(handler->GetSource()->GetOwner()))->GetType(), TYPESTR_PLUGIN)) {
          // The handler belongs to a plugin, so we need to tell the plugin object
          // to stop holding onto a pointer which will shortly be invalid
          ((Plugin *)(handler->GetSource()->GetOwner()))->RemoveOwnedEventHandler(handler);
        }
      }
    }
    if(handler->GetSink() != NULL) {
      // Check for other plugin-owned events
      if(!strcmp(((ebDynamicObject *)(handler->GetSink()))->GetType() , TYPESTR_PLUGIN)) {
        // The handler belongs to a plugin, so we need to tell the plugin object
        // to stop holding onto a pointer which will shortly be invalid
        ((Plugin *)(handler->GetSink()))->RemoveOwnedEventHandler(handler);
      }
    }
    delete handler;
  }
}

void ebDynamicObject::ClearMethodHandlers(void) {
  // Remove all attached method handlers
  while(methodhandlers.size() > 0) {
    MethodHandler *handler = methodhandlers.back();
    methodhandlers.pop_back();
    if(handler->GetOwner() != NULL) {
      if(!strcmp(handler->GetOwner()->GetType(), "CustomObject")) {
        if(!strcmp(((ebDynamicObject *)(handler->GetOwner()->GetOwner()))->GetType(), TYPESTR_PLUGIN)) {
          // The handler belongs to a plugin, so we need to tell the plugin object
          // to stop holding onto a pointer which will shortly be invalid
          ((Plugin *)(handler->GetOwner()->GetOwner()))->RemoveOwnedMethodHandler(handler);
        }
      }
    }
    delete handler;
  }
}

bool ebDynamicObject::IsFindable(void) {
  return findable;
}

void ebDynamicObject::SetContainer(NotifyObject *container) {
  if(this->container != NULL && container != NULL) {
    cerr << "ebDynamicObject::SetContainer - container already set!" << endl;
  }
  else
    this->container = container;
}

void ebDynamicObject::AddEvent(const char *eventname) {
  events.Add(eventname);
}

int ebDynamicObject::GetEventCount(void) {
  return events.size();
}

const char *ebDynamicObject::GetEvent(int index) {
  if(index >= 0 && index < events.size())
    return events[index];
  else
    return NULL;
}

const char *ebDynamicObject::GetDefaultEvent(void) {
  if(defaultevent > -1)
    return events[defaultevent];
  else
    return NULL;
}

