/***************************************************************************
                          scrollbarwidget.cpp  -  eboxy scrollbar widget
                             -------------------
    begin                : Sun Dec 1 2002
    copyright            : (C) 2002 by Paul Eggleton
    email                : bluelightning@bluelightning.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "gui_extensions.h"
#include "iostream"
#include "utils.h"
#include "main.h"

#include "scrollbarwidget.h"

using namespace std;

extern const char *TYPESTR_SCROLLBAR;

ScrollBarWidget::ScrollBarWidget(const char *name, int x, int y, int w, int h)
: ebWidget(name, x, y) {
  this->w = w;
  this->h = h;
  // Set defaults
  tabstop = true;
  value = 0;
  pagestep = 10;
  reversed = false;
  minimum = 0;
  maximum = 100;
  background = NULL;
  knob = NULL;
  knobfocus = NULL;
  backgroundstyle = IMAGESTYLE_NORMAL;
  
  this->type = TYPESTR_SCROLLBAR;

  AddProperty("background", ScrollBarWidget_GetBackground, ScrollBarWidget_SetBackground);
  AddProperty("backgroundstyle", ScrollBarWidget_GetBackgroundStyle, ScrollBarWidget_SetBackgroundStyle);
  AddProperty("knob", ScrollBarWidget_GetKnob, ScrollBarWidget_SetKnob);
  AddProperty("knobfocus", ScrollBarWidget_GetKnobFocus, ScrollBarWidget_SetKnobFocus);
  AddProperty("value", ScrollBarWidget_GetValue, ScrollBarWidget_SetValue);
  AddProperty("maximum", ScrollBarWidget_GetMaximum, ScrollBarWidget_SetMaximum);
  AddProperty("minimum", ScrollBarWidget_GetMinimum, ScrollBarWidget_SetMinimum);
  AddProperty("pagestep", ScrollBarWidget_GetPageStep, ScrollBarWidget_SetPageStep);
  AddProperty("reversed", ScrollBarWidget_GetReversed, ScrollBarWidget_SetReversed);

  FindProperty("width")->SetRequired(true);
  FindProperty("height")->SetRequired(true);

  AddEvent("OnChange");
  AddEvent("OnGotFocus");
  AddEvent("OnLostFocus");
}

GUI_Widget *ScrollBarWidget::Instantiate(void) {
  if(w<0) w=0;
  if(h<0) h=0;

  GUI_ExtScrollBar *scrollbar = new GUI_ExtScrollBar(name, x, y, w, h);

  if(this->background)
    scrollbar->SetBackgroundImage(this->background->surface);
/*  else {
    cerr << "Cannot instantiate scrollbar " << name << ": required property background is not set" << endl;
    return NULL;
  }*/
  scrollbar->SetBackgroundStyle(this->backgroundstyle);
  if(this->knob)
    scrollbar->SetKnobImage(this->knob->surface);
  if(this->knobfocus)
    scrollbar->SetKnobFocusImage(this->knobfocus->surface);
  scrollbar->SetMaximum(this->maximum);
  scrollbar->SetMinimum(this->minimum);
  scrollbar->SetValue(this->value);
  scrollbar->SetPageStep(this->pagestep);
  scrollbar->SetReversed(this->reversed);

  GUI_EventHandler<ScrollBarWidget> *cb = new GUI_EventHandler<ScrollBarWidget>(this, &ScrollBarWidget::OnChange);
  scrollbar->SetMovedCallback(cb);
	cb->DecRef();

  realwidget = scrollbar;
  return scrollbar;
}

void ScrollBarWidget::SetBackgroundImage(ImageEntry *image) {
  this->background = image;
}

void ScrollBarWidget::SetBackgroundStyle(int style) {
  this->backgroundstyle = style;
}

void ScrollBarWidget::SetKnobImage(ImageEntry *image) {
  this->knob = image;
}

void ScrollBarWidget::SetKnobFocusImage(ImageEntry *image) {
  this->knobfocus = image;
}

void ScrollBarWidget::SetValue(int value) {
  int oldvalue = this->value;

  if(value >= minimum && value <= maximum) {
    if(realwidget) {
      ((GUI_ExtScrollBar *)realwidget)->SetValue(value);
    }
    this->value = value;
    if(value != oldvalue)
      FireEvent("OnChange");
  }
}

void ScrollBarWidget::SetMaximum(int maximum) {
  if(maximum > minimum) {
    if(maximum < value) {
      SetValue(maximum);
    }
    this->maximum = maximum;
    if(realwidget)
      ((GUI_ExtScrollBar *)realwidget)->SetMaximum(maximum);
  }
}

void ScrollBarWidget::SetMinimum(int minimum) {
  if(minimum < maximum) {
    if(minimum > value) {
      SetValue(minimum);
    }
    this->minimum = minimum;
    if(realwidget)
      ((GUI_ExtScrollBar *)realwidget)->SetMinimum(minimum);
  }
}

void ScrollBarWidget::SetPageStep(int pagestep) {
  this->pagestep = pagestep;
  if(realwidget)
    ((GUI_ExtScrollBar *)realwidget)->SetPageStep(pagestep);
}

void ScrollBarWidget::SetReversed(bool reversed) {
  this->reversed = reversed;
  if(realwidget)
    ((GUI_ExtScrollBar *)realwidget)->SetReversed(reversed);
}

void ScrollBarWidget::Increment(void) {
  int newvalue = value + 1;

  if(newvalue <= maximum)
    SetValue(newvalue);
  else
    SetValue(maximum);

}

void ScrollBarWidget::Decrement(void) {
  int newvalue = value - 1;

  if(newvalue >= minimum)
    SetValue(newvalue);
  else
    SetValue(minimum);
}

void ScrollBarWidget::IncrementPage(void) {
  int newvalue = value + pagestep;

  if(newvalue <= maximum)
    SetValue(newvalue);
  else
    SetValue(maximum);
}

void ScrollBarWidget::DecrementPage(void) {
  int newvalue = value - pagestep;

  if(newvalue >= minimum)
    SetValue(newvalue);
  else
    SetValue(minimum);
}

void ScrollBarWidget::GotoStart(void) {
  SetValue(minimum);
}

void ScrollBarWidget::GotoEnd(void) {
  SetValue(maximum);
}

void ScrollBarWidget::GotFocus(void) {
  if(!focused) {
    this->FireEvent("OnGotFocus");
    ebWidget::GotFocus();
    ((GUI_ExtScrollBar *)realwidget)->MarkChanged();
  }
}

void ScrollBarWidget::LostFocus(void) {
  if(focused) {
    this->FireEvent("OnLostFocus");
    ebWidget::LostFocus();
    ((GUI_ExtScrollBar *)realwidget)->MarkChanged();
  }
}

void ScrollBarWidget::OnChange(GUI_Object *sender) {
  int currvalue = ((GUI_ExtScrollBar *)realwidget)->GetValue();

  if(currvalue != value) {
    value = currvalue;
    FireEvent("OnChange");
  }
}


/* Property get/set functions */

// -- background --

const char *ScrollBarWidget_GetBackground(ebCustomObject *sender) {
  if( ((ScrollBarWidget *)sender)->background )
    return ((ScrollBarWidget *)sender)->background->filename;
  else
    return "";
}

void ScrollBarWidget_SetBackground(ebCustomObject *sender, const char *value) {
  ImageEntry *image = loadImage(value);
  if(image)
    ((ScrollBarWidget *)sender)->SetBackgroundImage(image);
}

// -- backgroundstyle --

const char *ScrollBarWidget_GetBackgroundStyle(ebCustomObject *sender) {
  return styleinttostr( ((ScrollBarWidget *)sender)->backgroundstyle );
}

void ScrollBarWidget_SetBackgroundStyle(ebCustomObject *sender, const char *value) {
  ((ScrollBarWidget *)sender)->SetBackgroundStyle(stylestrtoint(value));
}

// -- knob --

const char *ScrollBarWidget_GetKnob(ebCustomObject *sender) {
  if( ((ScrollBarWidget *)sender)->knob )
    return ((ScrollBarWidget *)sender)->knob->filename;
  else
    return "";
}

void ScrollBarWidget_SetKnob(ebCustomObject *sender, const char *value) {
  ImageEntry *image = loadImage(value);
  if(image)
    ((ScrollBarWidget *)sender)->SetKnobImage(image);
}

// -- knobfocus --

const char *ScrollBarWidget_GetKnobFocus(ebCustomObject *sender) {
  if( ((ScrollBarWidget *)sender)->knobfocus )
    return ((ScrollBarWidget *)sender)->knobfocus->filename;
  else
    return "";
}

void ScrollBarWidget_SetKnobFocus(ebCustomObject *sender, const char *value) {
  ImageEntry *image = loadImage(value);
  if(image)
    ((ScrollBarWidget *)sender)->SetKnobFocusImage(image);
}

// -- value --

const char *ScrollBarWidget_GetValue(ebCustomObject *sender) {
  return(inttostr( ((ScrollBarWidget *)sender)->value ));
}

void ScrollBarWidget_SetValue(ebCustomObject *sender, const char *value) {
  ((ScrollBarWidget *)sender)->SetValue(atoi(value));
}

// -- maximum --

const char *ScrollBarWidget_GetMaximum(ebCustomObject *sender) {
  return(inttostr( ((ScrollBarWidget *)sender)->maximum ));
}

void ScrollBarWidget_SetMaximum(ebCustomObject *sender, const char *value) {
  ((ScrollBarWidget *)sender)->SetMaximum(atoi(value));
}

// -- minimum --

const char *ScrollBarWidget_GetMinimum(ebCustomObject *sender) {
  return(inttostr( ((ScrollBarWidget *)sender)->minimum ));
}

void ScrollBarWidget_SetMinimum(ebCustomObject *sender, const char *value) {
  ((ScrollBarWidget *)sender)->SetMinimum(atoi(value));
}

// -- pagestep --

const char *ScrollBarWidget_GetPageStep(ebCustomObject *sender) {
  return(inttostr( ((ScrollBarWidget *)sender)->pagestep ));
}

void ScrollBarWidget_SetPageStep(ebCustomObject *sender, const char *value) {
  ((ScrollBarWidget *)sender)->SetPageStep(atoi(value));
}

// -- reversed --

const char *ScrollBarWidget_GetReversed(ebCustomObject *sender) {
  return(booltostr( ((ScrollBarWidget *)sender)->reversed ));
}

void ScrollBarWidget_SetReversed(ebCustomObject *sender, const char *value) {
  ((ScrollBarWidget *)sender)->SetReversed(strtobool(value));
}
