/***************************************************************************
                          skinparser.cpp  -  XML skin file parser
                             -------------------
    begin                : Thu Nov 7 2002
    copyright            : (C) 2002 by Paul Eggleton
    email                : bluelightning@bluelightning.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <cstdio>
#include <iostream>
#include <cstring>
#include <cstdlib>
#include <vector>

#include "libxml/xmlmemory.h"
#include "libxml/parser.h"
#include "libxml/tree.h"
#include "unistd.h"

#include "widgets.h"
#include "utils.h"
#include "errors.h"

#include "buttonwidget.h"
#include "picturewidget.h"
#include "togglebuttonwidget.h"
#include "labelwidget.h"
#include "scrollbarwidget.h"
#include "textfieldwidget.h"
#include "listboxwidget.h"
#include "page.h"
#include "timer.h"
#include "systemobject.h"
#include "pluginmanager.h"
#include "scripteventhandler.h"
#include "main.h"

#include "skinparser.h"

using namespace std;

extern vector<ebPage *> pages;
extern SystemObject *systemObject;
extern PluginManager plugin_manager;
extern const char *DTD_FILE;
extern const int DEFAULT_FONTSIZE;
extern const SDL_Color DEFAULT_FONT_COLOR;

extern const char *TYPESTR_LABEL;
extern const char *TYPESTR_BUTTON;
extern const char *TYPESTR_PICTURE;
extern const char *TYPESTR_LISTBOX;
extern const char *TYPESTR_TEXTFIELD;
extern const char *TYPESTR_SCROLLBAR;
extern const char *TYPESTR_TOGGLEBUTTON;

vector<GenericTemplate *> templates;

int parseSkin(char *filename) {
  xmlDocPtr doc;
  xmlNodePtr cur;
  int rc = 0;

  if(!checkFileExists(filename))
    return EBERR_FILE_ERROR;

  // Build a DOM tree from the XML file
  xmlLineNumbersDefault(1);
  doc = xmlParseFile(filename);
  if (doc == NULL) {
    cerr << "Failed to parse XML file" << endl;
    return EBERR_XML_PARSE_FAILED;
  }

  // Change to the directory where the XML file is (for image/font paths)
  char *filepath = getFilePath(filename);
  chdir(filepath);

  // Check the document "type"
  cur = xmlDocGetRootElement(doc);
  if (cur == NULL) {
    cerr << "XML file " << filename << " is empty" << endl;
    xmlFreeDoc(doc);
  	return EBERR_XML_EMPTY;
  }
  if (xmlStrcmp(cur->name, (const xmlChar *) "eboxy")) {
    cerr << "XML file " << filename << " has the wrong type (root node should be \'eboxy\')" << endl;
  	xmlFreeDoc(doc);
  	return EBERR_XML_NOT_VALID;
  }

  // Verify the XML file against the DTD
  xmlDtdPtr dtd = xmlParseDTD(NULL, (const xmlChar *) DTD_FILE);
  if (dtd == NULL) {
    xmlGenericError(xmlGenericErrorContext, "Could not parse DTD %s\n", DTD_FILE);
    return EBERR_DTD_PARSE_FAILED;
  } else {
    xmlValidCtxt cvp;
    cvp.userData = (void *) stderr;
    // FIXME: this doesn't print out the XML line number the error occurred in
    cvp.error = (xmlValidityErrorFunc) fprintf;
    cvp.warning = (xmlValidityWarningFunc) fprintf;
    if (!xmlValidateDtd(&cvp, doc, dtd)) {
      xmlGenericError(xmlGenericErrorContext, "XML file %s does not validate against DTD %s\n", filename, DTD_FILE);
      return EBERR_XML_NOT_VALID;
    }
  }

  // Walk the tree
  cur = cur->xmlChildrenNode;
  while(cur) {
    if(!xmlIsBlankNode(cur)) {
      if (!(xmlStrcmp(cur->name, (const xmlChar *) "templates")) ) {
        rc = parseTemplates(doc, cur);
        if(rc != EBERR_SUCCESS) {
          // Error occurred, message has already been displayed
          xmlFreeDoc(doc);
          return rc;
        }
      }
      if (!(xmlStrcmp(cur->name, (const xmlChar *) "pages")) ) {
        rc = parsePages(doc, cur);
        if(rc != EBERR_SUCCESS) {
          // Error occurred, message has already been displayed
          xmlFreeDoc(doc);
          return rc;
        }
      }
      if (!(xmlStrcmp(cur->name, (const xmlChar *) "system")) ) {
        rc = parseSystem(doc, cur);
        if(rc != EBERR_SUCCESS) {
          // Error occurred, message has already been displayed
          xmlFreeDoc(doc);
          return rc;
        }
      }
    }
    cur = cur->next;
  }

  // Done parsing
  xmlFreeDoc(doc);

  // See what we've got
  if(pages.size() == 0) {
    cerr << "XML file " << filename << " does not contain any pages" << endl;
    return EBERR_XML_NOPAGES;
  }

  return EBERR_SUCCESS;
}

int parsePages(xmlDocPtr doc, xmlNodePtr node) {
  xmlNodePtr cur;

  cur = node->xmlChildrenNode;
  while (cur != NULL) {
    if (!xmlStrcmp(cur->name, (const xmlChar *) "page")) {
      ebPage *page = parsePage(doc, cur);
      if(page)
        pages.push_back(page);
      else
        return EBERR_XML_BAD;
    }
    cur = cur->next;
  }
  return EBERR_SUCCESS;
}

int parseTemplates(xmlDocPtr doc, xmlNodePtr node) {
  xmlNodePtr cur;

  cur = node->xmlChildrenNode;
  while (cur != NULL) {
    // New parsing method
    if(cur->type == XML_ELEMENT_NODE) {
      GenericTemplate *gtemplate = parseTemplate(doc, cur);
      if(gtemplate)
        templates.push_back(gtemplate);
      else
        return EBERR_XML_BAD;
    }
    cur = cur->next;
  }
  return EBERR_SUCCESS;
}

int parseSystem(xmlDocPtr doc, xmlNodePtr node) {
  xmlNodePtr cur;

  cur = node->xmlChildrenNode;
  while (cur != NULL) {
    if ((!xmlStrcmp(cur->name, (const xmlChar *) "event")) ) {
      parseEvent(doc, cur, systemObject);
    }
    else if ((!xmlStrcmp(cur->name, (const xmlChar *) "timer")) ) {
      TimerObject *timer = parseTimer(doc, cur);
      if(timer)
        systemObject->AddObject(timer);
    }
    else if ((!xmlStrcmp(cur->name, (const xmlChar *) "plugin")) ) {
      Plugin *plugin = parsePlugin(doc, cur);
      if(plugin)
        systemObject->AddUnownedObject(plugin); // plugin is managed automatically by the plugin manager
    }
    cur = cur->next;
  }
  return EBERR_SUCCESS;
}

ebPage *parsePage(xmlDocPtr doc, xmlNodePtr node) {
  xmlNodePtr cur;
  char *str;
  ebPage *page;
  ImageEntry *background;
  char *name;

  str = (char *)xmlGetProp(node, (const xmlChar *) "name");
  if(str) {
    if(!checkName(str)) {
      cerr << "\'" << str << "\' is not a valid name (line " << xmlGetLineNo(node) << ")" << endl;
      free(str);
      return NULL;
    }
    name = strdup(str);
    free(str);
  }

  str = (char *)xmlGetProp(node, (const xmlChar *) "background");
  if(str) {
    background = loadImage(str);
    free(str);
    if(background == NULL) {
      // Error occurred, and message has already been displayed
      return NULL;
    }
  }
  else
    background = NULL;

  page = new ebPage(name, 0, 0, 10, 10); // dummy values
  page->SetBackground(background);
  free(name); // important

  cur = node->xmlChildrenNode;
  while (cur != NULL) {
    if(cur->type == XML_ELEMENT_NODE) {
      if ((!xmlStrcmp(cur->name, (const xmlChar *) "plugin")) ) {
        Plugin *plugin = parsePlugin(doc, cur);
        if(plugin)
          page->AddUnownedObject(plugin); // plugin is managed automatically by the plugin manager
      }
      else if ((!xmlStrcmp(cur->name, (const xmlChar *) "event")) ) {
        parseEvent(doc, cur, page);
      }
      else if ((!xmlStrcmp(cur->name, (const xmlChar *) "timer")) ) {
        TimerObject *timer = parseTimer(doc, cur);
        if(timer)
          page->AddObject(timer);
        else
          return NULL;
      }
      else {
        // A widget
        ebWidget *widget = parseWidget(doc, cur);
        if(widget)
          page->AddWidget(widget);
        else
          return NULL;
      }
    }
    cur = cur->next;
  }

  return page;
}

ebWidget *parseWidget(xmlDocPtr doc, xmlNodePtr node) {
  xmlNodePtr cur;
  char *str;
  char *name;
  ebWidget *widget;
  GenericTemplate *gtemplate = NULL;
  static int count = 0;
  
  // Grab name if specified, otherwise make one up (useful for debugging)
  str = (char *)xmlGetProp(node, (const xmlChar *) "name");
  if(str) {
    if(!checkName(str)) {
      cerr << "\'" << str << "\' is not a valid name (line " << xmlGetLineNo(node) << ")" << endl;
      free(str);
      return NULL;
    }
    name = strdup(str);
    free(str);
  }
  else {
    name = newstring(strlen((const char *)node->name) + 6);
    sprintf(name, "_%s%d", (const char *)node->name, ++count);
  }

  widget = CreateWidget(name, (const char *)node->name);
  
  // Grab template
  str = (char *)xmlGetProp(node, (const xmlChar *) "template");
  if(str) {
    gtemplate = findTemplate(str);
    if(!gtemplate) {
      cerr << "Non-existent template " << str << " specified (line " << xmlGetLineNo(node) << ")" << endl;
      delete widget;
      free(name);
      free(str);
      return NULL;
    }
    free(str);
  }

  // Set properties
  int propcount = widget->GetPropertyCount();

  int index;
  CustomProperty *prop;
  bool propset;
  for(int i=0;i<propcount;i++) {
    prop = widget->GetProperty(i);
    if(!(prop->IsReadOnly())) {
      propset = false;
      str = (char *)xmlGetProp(node, (const xmlChar *) prop->GetName());
      if(str) {
        try {
          prop->SetValue(str);
        }
        catch(int e) {
          // Unable to set property value
          delete widget;
          free(name);
          return NULL;
        }
        propset = true;
        free(str);
      }
      else {
        // Check for template value
        if(gtemplate) {
          index = gtemplate->attributes.IndexOfName(prop->GetName());
          if(index > -1) {
            try {
              prop->SetValue(gtemplate->attributes.GetValue(index));
            }
            catch(int e) {
              // Unable to set property value
              delete widget;
              free(name);
              return NULL;
            }
            propset = true;
          }
        }
      }

      if(!propset && prop->IsRequired()) {
        // Property required and not set
        cerr << "Required property " << prop->GetName() << " of object " << name << " has not been specified (line " << xmlGetLineNo(node) << ")" << endl;
        delete widget;
        free(name);
        return NULL;
      }
    }
  }

  if(widget->GetDefaultEvent()) {
    // Grab default event
    str = (char *)xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
    if(str != NULL) {
      if(!isEmptyString(str)) {
        Script *script = new Script("noname", str); // FIXME: name
        script->SetLineOffset(xmlGetLineNo(node));
        widget->RegisterEventHandler(widget->GetDefaultEvent(), script); // FIXME: check if default event exists
      }
      free(str);
    }
  }
  else if(widget->GetDefaultProperty()) {
    // Grab default property (currently only used by labels)
    str = (char *)xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
    if(str != NULL) {
      widget->SetPropertyValue(widget->GetDefaultProperty(), str);
      free(str);
    }
  }

  // Grab attached events
  cur = node->xmlChildrenNode;
  while (cur != NULL) {
    if (!xmlStrcmp(cur->name, (const xmlChar *) "event")) {
      parseEvent(doc, cur, widget);
    }
    cur = cur->next;
  }

  free(name);
  return widget;
}

void parseEvent(xmlDocPtr doc, xmlNodePtr node, ebDynamicObject *parent) {
  EventTemplate *evt = parseEvent(doc, node);

  if(evt->script && evt->type) {
    EventHandler *handler = parent->RegisterEventHandler(evt->type, evt->script);
    handler->SetThreaded(evt->threaded);
  }
  else {
    cerr << "Invalid event specified for " << parent->GetName() << " (line " << xmlGetLineNo(node) << ")" << endl;
    if(evt->script)
      delete evt->script;
  }

  if(evt->type)
    free(evt->type);
  delete evt;
  // We don't delete evt->script here because it is being used (or already deleted on error) - see ScriptEventHandler
}

EventTemplate *parseEvent(xmlDocPtr doc, xmlNodePtr node) {
  EventTemplate *evt;
  char *str;

  evt = new EventTemplate;

  // Grab event type
  evt->type = (char *)xmlGetProp(node, (const xmlChar *) "type");

  str = (char *)xmlGetProp(node, (const xmlChar *) "wait");
  if(str) {
    evt->threaded = !strtobool(str);
    free(str);
  }
  else
    evt->threaded = true;

  // Grab script
  str = (char *)xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
  if(str != NULL && !isEmptyString(str)) {
    evt->script = new Script("script1", str);
    evt->script->SetLineOffset(xmlGetLineNo(node));
  }
  else
    evt->script = NULL;
  if(str)
    free(str);

  return evt;
}

TimerObject *parseTimer(xmlDocPtr doc, xmlNodePtr node) {
  char *str;
  char *name;
  int interval = 1000;
  bool autostart = false;
  TimerObject *timer;
  static int count = 0;

  // Grab name if specified, otherwise make one up (useful for debugging)
  str = (char *)xmlGetProp(node, (const xmlChar *) "name");
  if(str) {
    if(!checkName(str)) {
      cerr << "\'" << str << "\' is not a valid name (line " << xmlGetLineNo(node) << ")" << endl;
      free(str);
      return NULL;
    }
    name = strdup(str);
    free(str);
  }
  else {
    name = newstring(15);
    sprintf(name, "_timer%d", ++count);
  }

  str = (char *)xmlGetProp(node, (const xmlChar *) "interval");
  if(str) {
    interval = atoi(str);
    free(str);
  }

  str = (char *)xmlGetProp(node, (const xmlChar *) "enabled");
  if(str) {
    autostart = strtobool(str);
    free(str);
  }

  // Create the timer
  timer = new TimerObject(name, interval, autostart);

  // Grab attached event handlers
  str = (char *)xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
  if(str != NULL) {
    if(strlen(str) > 0) {
      Script *script = new Script("noname", str); // FIXME: name
      script->SetLineOffset(xmlGetLineNo(node));
      timer->RegisterEventHandler("OnTimer", script);
    }
    free(str);
  }

  free(name);
  return timer;
}

Plugin *parsePlugin(xmlDocPtr doc, xmlNodePtr node) {
  char *str;
  xmlNodePtr cur, cur2;
  Plugin *plugin;
  EventTemplate *evt;
  ScriptEventHandler *handler;
  char *objectname;

  str = (char *)xmlGetProp(node, (const xmlChar *) "module");
  if(str) {
    // Load the plugin
    if(plugin_manager.LoadPlugin(str, false, &plugin) != EBERR_SUCCESS) {
      // Error occurred during load (we assume that an error message has been displayed)
      free(str);
      return NULL;
    }
    free(str);
  }
  else {
    // DTD shouldn't allow this, but what the heck...
    cerr << "No plugin module specified (line " << xmlGetLineNo(node) << ")" << endl;
    return NULL;
  }

  // Parse contents of plugin object element (currently, only attached events)
  cur = node->xmlChildrenNode;
  while (cur != NULL) {
    if (!xmlStrcmp(cur->name, (const xmlChar *) "pluginobject")) {
      // Get the name of the object we're looking at
      objectname = (char *)xmlGetProp(cur, (const xmlChar *) "name");
      // Get events that the user wants to bind to this object
      cur2 = cur->xmlChildrenNode;
      while (cur2 != NULL) {
        if (!xmlStrcmp(cur2->name, (const xmlChar *) "event")) {
          // Register an event (well, actually the registering gets done later)
          evt = parseEvent(doc, cur2);
          evt->sourceobject = strdup(objectname);
          plugin->AddEventHandler(evt);
        }
        cur2 = cur2->next;
      }
      // Release object name string
      free(objectname);
    }
    cur = cur->next;
  }

  return plugin;
}

void clearTemplates(void) {
  while(templates.size() > 0) {
    GenericTemplate *templt = templates.back();
    templates.pop_back();
    if(templt->name)
      free(templt->name);
    delete templt;
  }  
}

GenericTemplate *parseTemplate(xmlDocPtr doc, xmlNodePtr node) {
  GenericTemplate *gtemplate;

  // Create template object
  gtemplate = new GenericTemplate;
  
  // Set template type
  if (!xmlStrcmp(node->name, (const xmlChar *) "buttontemplate"))
    gtemplate->type = TT_BUTTON;
  else if (!xmlStrcmp(node->name, (const xmlChar *) "labeltemplate"))
    gtemplate->type = TT_LABEL;
  else if (!xmlStrcmp(node->name, (const xmlChar *) "listboxtemplate"))
    gtemplate->type = TT_LISTBOX;
  else if (!xmlStrcmp(node->name, (const xmlChar *) "scrollbartemplate"))
    gtemplate->type = TT_SCROLLBAR;
  else if (!xmlStrcmp(node->name, (const xmlChar *) "textfieldtemplate"))
    gtemplate->type = TT_TEXTFIELD;
  else if (!xmlStrcmp(node->name, (const xmlChar *) "togglebuttontemplate"))
    gtemplate->type = TT_TOGGLEBUTTON;
  else {
    cerr << "parseTemplate: Unknown template type '" << node->name << "'" << endl;
    delete gtemplate;
    return NULL;
  }
  
  // Get attributes into a StringList
  _xmlAttr *prop = node->properties;
  while(prop) {
    if(strcmp((char *)prop->name, "name") == 0)
      gtemplate->name = strdup((char *)prop->children->content);
    else
      gtemplate->attributes.SetValue((char *)prop->name, (char *)prop->children->content);
    prop = prop->next;
  }

  return gtemplate;
}

GenericTemplate *findTemplate(const char *templatename) {
  // Find the referenced template
  for(int i=0;i<templates.size();i++) {
    if(!strcasecmp(templates[i]->name, templatename)) {
      return templates[i];
    }
  }
  return NULL;
}

int applyTemplate(ebWidget *widget, const char *templatename) {
  // Find specified template
  GenericTemplate *gtemplate = findTemplate(templatename);
  if(!gtemplate)
    return EBERR_OBJECT_NOT_FOUND;
  
  // Set widget properties from template
  int propcount = widget->GetPropertyCount();
  for(int i=0;i<propcount;i++) {
    CustomProperty *prop = widget->GetProperty(i);
    int index = gtemplate->attributes.IndexOfName(prop->GetName());
    if(index > -1) {
      try {
        prop->SetValue(gtemplate->attributes.GetValue(index));
      }
      catch(int e) {
        // Failed to set property
        return e;
      }
    }
  }

  return EBERR_SUCCESS;
}

ebWidget *CreateWidget(const char *name, const char *type) {
  if(!strcmp(type, TYPESTR_BUTTON))
    return new ButtonWidget(name, 0, 0, -1, -1);
  else if(!strcmp(type, TYPESTR_LABEL))
    return new LabelWidget(name, 0, 0, 0, 0, "", NULL);
  else if(!strcmp(type, TYPESTR_PICTURE))
    return new PictureWidget(name, 0, 0, NULL);
  else if(!strcmp(type, TYPESTR_LISTBOX))
    return new ListBoxWidget(name, 0, 0, -1, -1, NULL, DEFAULT_FONT_COLOR);
  else if(!strcmp(type, TYPESTR_TOGGLEBUTTON))
    return new ToggleButtonWidget(name, 0, 0, -1, -1);
  else if(!strcmp(type, TYPESTR_SCROLLBAR))
    return new ScrollBarWidget(name, 0, 0, -1, -1);
  else if(!strcmp(type, TYPESTR_TEXTFIELD))
    return new TextFieldWidget(name, 0, 0, -1, -1, NULL, DEFAULT_FONT_COLOR, 10);
  else {
    cerr << "CreateWidget: Unknown widget type '" << type << "'" << endl;
    return NULL;
  }
}
