/***************************************************************************
                          textfieldwidget.cpp  -  eboxy textfield widget
                             -------------------
    begin                : Sun Dec 8 2002
    copyright            : (C) 2002 by Paul Eggleton
    email                : bluelightning@bluelightning.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <iostream>
#include <cstring>
#include <cstdlib>

#include "gui_extensions.h"
#include "utils.h"
#include "main.h"

#include "textfieldwidget.h"

// Prototype for timer callback function
static Uint32 timer_callback(Uint32 interval, void *param);

using namespace std;

extern const char *TYPESTR_TEXTFIELD;
extern const int DEFAULT_FONTSIZE;

// This array defines what characters are available using keypad (digit) text input
const char *digitchars[] = {
  "0 -+",
  "1.,:;\'\"",
  "abc2",
  "def3",
  "ghi4",
  "jkl5",
  "mno6",
  "pqrs7",
  "tuv8",
  "wxyz9",
  NULL  // end of list
};

TextFieldWidget::TextFieldWidget(const char *name, int x, int y, int w, int h, FontEntry *font, SDL_Color fontcolor, int maxlength)
: ebWidget(name, x, y) {
  this->w = w;
  this->h = h;
  this->font = font;
  this->fontsize = DEFAULT_FONTSIZE;
  this->fontcolor = fontcolor;
  this->maxlength = maxlength;

  this->inpdigit = -1;
  this->inpdigitptr = NULL;

  // Set defaults
  this->text = strdup("");

  this->tabstop = true;
  this->type = TYPESTR_TEXTFIELD;
  this->background = NULL;
  this->backgroundfocus = NULL;
  this->backgroundstyle = IMAGESTYLE_NORMAL;
  this->borderx = 0;
  this->bordery = 0;

  AddProperty("text", TextFieldWidget_GetText, TextFieldWidget_SetText);
  AddProperty("font", TextFieldWidget_GetFont, TextFieldWidget_SetFont, true);
  AddProperty("fontsize", TextFieldWidget_GetFontSize, TextFieldWidget_SetFontSize);
  AddProperty("fontcolor", TextFieldWidget_GetFontColor, TextFieldWidget_SetFontColor);
  AddProperty("background", TextFieldWidget_GetBackground, TextFieldWidget_SetBackground);
  AddProperty("backgroundfocus", TextFieldWidget_GetBackgroundFocus, TextFieldWidget_SetBackgroundFocus);
  AddProperty("backgroundstyle", TextFieldWidget_GetBackgroundStyle, TextFieldWidget_SetBackgroundStyle);
  AddProperty("maxlength", TextFieldWidget_GetMaxLength, TextFieldWidget_SetMaxLength);
  AddProperty("borderx", TextFieldWidget_GetBorderX, TextFieldWidget_SetBorderX);
  AddProperty("bordery", TextFieldWidget_GetBorderY, TextFieldWidget_SetBorderY);
  AddProperty("cursorpos", TextFieldWidget_GetCursorPos, TextFieldWidget_SetCursorPos);

  FindProperty("width")->SetRequired(true);
  FindProperty("height")->SetRequired(true);

  AddEvent("OnChange");
  AddEvent("OnGotFocus");
  AddEvent("OnLostFocus");
}

TextFieldWidget::~TextFieldWidget(void) {
  free(text);
}

GUI_Widget *TextFieldWidget::Instantiate(void) {
  if(w<0) w=0;
  if(h<0) h=0;

  GUI_TextField *textentry = new GUI_TextField(name, x, y, w, h, font->font, maxlength);

  if(this->background)
    textentry->SetNormalImage(this->background->surface);
  if(this->backgroundfocus)
    textentry->SetFocusImage(this->backgroundfocus->surface);
  textentry->SetBackgroundStyle(this->backgroundstyle);
  textentry->SetBorderX(borderx);
  textentry->SetBorderY(bordery);
  textentry->SetTextColor(fontcolor.r, fontcolor.g, fontcolor.b);
  textentry->SetText(text);

  GUI_EventHandler<TextFieldWidget> *cb = new GUI_EventHandler<TextFieldWidget>(this, &TextFieldWidget::OnChange);
  textentry->SetChangedCallback(cb);
	cb->DecRef();

  inpdigit = -1;

  realwidget = textentry;
  return textentry;
}

void TextFieldWidget::Uninstantiate(void) {
  // Stop timer if running
  if(inpdigit != -1)
    SDL_RemoveTimer(inpdigit_timer);

  ebWidget::Uninstantiate();
}

void TextFieldWidget::SetBackgroundImage(ImageEntry *image) {
  this->background = image;
}

void TextFieldWidget::SetBackgroundFocusImage(ImageEntry *image) {
  this->backgroundfocus = image;
}

void TextFieldWidget::SetBackgroundStyle(int style) {
  this->backgroundstyle = style;
}

void TextFieldWidget::SetFont(FontEntry *font) {
  this->font = font;
  if(realwidget) {
    ((GUI_TextField *)realwidget)->SetFont(this->font->font);
    realwidget->MarkChanged();
  }
}

void TextFieldWidget::SetFontColor(SDL_Color c) {
  this->fontcolor = c;
  if(realwidget) {
    ((GUI_TextField *)realwidget)->SetTextColor(c.r, c.g, c.b);
    realwidget->MarkChanged();
  }
}

void TextFieldWidget::SetBorderX(int pixels) {
  this->borderx = pixels;
  if(realwidget)
    ((GUI_TextField *)realwidget)->SetBorderX(pixels);
}

void TextFieldWidget::SetBorderY(int pixels) {
  this->bordery = pixels;
  if(realwidget)
    ((GUI_TextField *)realwidget)->SetBorderY(pixels);
}

void TextFieldWidget::SendEvent(SDL_Event e) {
  if(realwidget) {
/*    SDL_Event e;
    e.type = SDL_KEYUP;
    e.key.type = SDL_KEYUP;
    e.key.state = SDL_PRESSED;
    e.key.keysym.key = key;
    e.key.keysym.mod = KMOD_NONE;
    e.key.keysym.scancode = 0;
    e.key.keysym.unicode = 0;*/
    ((GUI_TextField *)realwidget)->Event(&e, 0, 0);
  }
}

void TextFieldWidget::SendChar(char c) {
  if(realwidget)
    ((GUI_TextField *)realwidget)->SendChar(c);
}

void TextFieldWidget::Backspace(void) {
  if(realwidget)
    ((GUI_TextField *)realwidget)->Backspace();
}

void TextFieldWidget::DeleteCurrChar(void) {
  if(realwidget)
    ((GUI_TextField *)realwidget)->DeleteCurrChar();
}

void TextFieldWidget::SetText(const char *str) {
  free(this->text);
  this->text = strdup(str);
  if(strlen(str) > maxlength)
    this->text[maxlength] = '\0';
  if(realwidget) {
    ((GUI_TextField *)realwidget)->SetText(str);
    SetCursorPos(strlen(this->text));
  }
}

const char *TextFieldWidget::GetText(void) {
  return text;
}

int TextFieldWidget::GetCursorPos(void) {
  if(realwidget) {
    return ((GUI_TextField *)realwidget)->GetCursorPos();
  }
  return 0;
}

void TextFieldWidget::SetCursorPos(int pos) {
  if(realwidget) {
    ((GUI_TextField *)realwidget)->SetCursorPos(pos);
  }
}

// FIXME not i18n compatible
void TextFieldWidget::InputDigit(int chr) {
  if(inpdigit > -1 && inpdigit == chr) {
    // We are already in digit input mode and the same button was pressed
    inpdigitptr++;
    if(inpdigitptr[0] == '\0')
      inpdigitptr = digitchars[inpdigit];  // back to start
    Backspace();
    SendChar(inpdigitptr[0]);
    // Reset timer
    SDL_RemoveTimer(inpdigit_timer);
    inpdigit_timer = SDL_AddTimer(1000, &timer_callback, (void *)this);
  }
  else {
    // New key
    // Stop timer if running
    if(inpdigit != -1)
      SDL_RemoveTimer(inpdigit_timer);

    inpdigit = chr;
    if(inpdigit != -1) {
      inpdigitptr = digitchars[inpdigit];  // back to start
      SendChar(inpdigitptr[0]);
      // Start timer
      inpdigit_timer = SDL_AddTimer(1000, &timer_callback, (void *)this);
    }
  }
}

void TextFieldWidget::GotFocus(void) {
  if(!focused) {
    this->FireEvent("OnGotFocus");
    ebWidget::GotFocus();
    ((GUI_TextField *)realwidget)->MarkChanged();
  }
}

void TextFieldWidget::LostFocus(void) {
  if(focused) {
    this->FireEvent("OnLostFocus");
    ebWidget::LostFocus();
    ((GUI_TextField *)realwidget)->MarkChanged();
  }
}

void TextFieldWidget::OnChange(GUI_Object *sender) {
  free(text);
  text = strdup(((GUI_TextField *)realwidget)->GetText());
  FireEvent("OnChange");
}

static Uint32 timer_callback(Uint32 interval, void *param) {
  // Reset digit input
  ((TextFieldWidget *)param)->InputDigit(-1);
  return 0;
}


/* Property get/set functions */

// -- text --

const char *TextFieldWidget_GetText(ebCustomObject *sender) {
  return ((TextFieldWidget *)sender)->text;
}

void TextFieldWidget_SetText(ebCustomObject *sender, const char *value) {
  ((TextFieldWidget *)sender)->SetText(value);
}

// -- font --

const char *TextFieldWidget_GetFont(ebCustomObject *sender) {
  return ((TextFieldWidget *)sender)->font->filename;
}

void TextFieldWidget_SetFont(ebCustomObject *sender, const char *value) {
  TextFieldWidget *textfield = ((TextFieldWidget *)sender);

  FontEntry *font = loadFont((char *)value, textfield->fontsize);
  if(font)
    textfield->SetFont(font);
}

// -- fontsize --

const char *TextFieldWidget_GetFontSize(ebCustomObject *sender) {
  return inttostr(((TextFieldWidget *)sender)->fontsize);
}

void TextFieldWidget_SetFontSize(ebCustomObject *sender, const char *value) {
  TextFieldWidget *textfield = ((TextFieldWidget *)sender);
  int size = atoi(value);

  textfield->fontsize = size;
  if(textfield->font)
    TextFieldWidget_SetFont(textfield, textfield->font->filename);
}

// -- fontcolor --

const char *TextFieldWidget_GetFontColor(ebCustomObject *sender) {
  return colortostr( ((TextFieldWidget *)sender)->fontcolor );
}

void TextFieldWidget_SetFontColor(ebCustomObject *sender, const char *value) {
  ((TextFieldWidget *)sender)->SetFontColor(strtocolor((char *)value));
}

// -- background --

const char *TextFieldWidget_GetBackground(ebCustomObject *sender) {
  if( ((TextFieldWidget *)sender)->background )
    return ((TextFieldWidget *)sender)->background->filename;
  else
    return "";
}

void TextFieldWidget_SetBackground(ebCustomObject *sender, const char *value) {
  ImageEntry *image = loadImage(value);
  if(image)
    ((TextFieldWidget *)sender)->SetBackgroundImage(image);
}

// -- backgroundfocus --

const char *TextFieldWidget_GetBackgroundFocus(ebCustomObject *sender) {
  if( ((TextFieldWidget *)sender)->backgroundfocus )
    return ((TextFieldWidget *)sender)->backgroundfocus->filename;
  else
    return "";
}

void TextFieldWidget_SetBackgroundFocus(ebCustomObject *sender, const char *value) {
  ImageEntry *image = loadImage(value);
  if(image)
    ((TextFieldWidget *)sender)->SetBackgroundFocusImage(image);
}

// -- backgroundstyle --

const char *TextFieldWidget_GetBackgroundStyle(ebCustomObject *sender) {
  return styleinttostr( ((TextFieldWidget *)sender)->backgroundstyle );
}

void TextFieldWidget_SetBackgroundStyle(ebCustomObject *sender, const char *value) {
  ((TextFieldWidget *)sender)->SetBackgroundStyle(stylestrtoint(value));
}

// -- maxlength --

const char *TextFieldWidget_GetMaxLength(ebCustomObject *sender) {
  return(inttostr( ((TextFieldWidget *)sender)->maxlength ));
}

void TextFieldWidget_SetMaxLength(ebCustomObject *sender, const char *value) {
  if( ((TextFieldWidget *)sender)->realwidget == NULL )
    ((TextFieldWidget *)sender)->maxlength = atoi(value); // FIXME: dodgy
  else
    cerr << "SetPropertyValue: Property maxlength is read-only" << endl;
}

// -- borderx --

const char *TextFieldWidget_GetBorderX(ebCustomObject *sender) {
  return inttostr( ((TextFieldWidget *)sender)->borderx );
}

void TextFieldWidget_SetBorderX(ebCustomObject *sender, const char *value) {
  ((TextFieldWidget *)sender)->SetBorderX(atoi(value));
}

// -- bordery --

const char *TextFieldWidget_GetBorderY(ebCustomObject *sender) {
  return inttostr( ((TextFieldWidget *)sender)->bordery );
}

void TextFieldWidget_SetBorderY(ebCustomObject *sender, const char *value) {
  ((TextFieldWidget *)sender)->SetBorderY(atoi(value));
}

// -- cursorpos --

const char *TextFieldWidget_GetCursorPos(ebCustomObject *sender) {
  return inttostr( ((TextFieldWidget *)sender)->GetCursorPos() );
}

void TextFieldWidget_SetCursorPos(ebCustomObject *sender, const char *value) {
  ((TextFieldWidget *)sender)->SetCursorPos(atoi(value));
}

