/***************************************************************************
                      timer.cpp - object wrapper for SDL timer
                             -------------------
    begin                : Tue Nov 5 2002
    copyright            : (C) 2002 by Paul Eggleton
    email                : bluelightning@bluelightning.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "SDL.h"

#include "utils.h"

#include "timer.h"

using namespace std;

extern const char *TYPESTR_TIMER;

// Prototype of callback procedure
static Uint32 timer_callback(Uint32 interval, void *param);

TimerObject::TimerObject(const char *name, int interval, bool autostart)
: ebCustomObject(name) {
  this->interval = interval;
  this->enabled = false;
  this->autostart = autostart;
  this->type = TYPESTR_TIMER;

  AddProperty("enabled", TimerObject_GetEnabled, TimerObject_SetEnabled);
  AddProperty("interval", TimerObject_GetInterval, TimerObject_SetInterval);

  AddMethod("start", 0, TimerObject_Start);
  AddMethod("stop", 0, TimerObject_Stop);

  AddEvent("OnTimer");
  defaultevent = 0;
}

TimerObject::~TimerObject(void) {
  // Make sure timer isn't still going!
  SetEnabled(false);
}

void TimerObject::SetInterval(int interval) {
  if(interval != this->interval) {
    this->interval = interval;
    if(this->enabled) {
      // Restart timer so that change takes effect
      SetEnabled(false);
      SetEnabled(true);
    }
  }
}

void TimerObject::SetEnabled(bool enable) {
  if(enable != this->enabled) {
    if(enable) {
      // Start timer
      this->timer_id = SDL_AddTimer(interval, &timer_callback, (void *)this);
    }
    else {
      // Stop timer
      SDL_RemoveTimer(timer_id);
    }
    this->enabled = enable;
  }
}

void TimerObject::Fire(void) {
  FireEvent("OnTimer");
}

void TimerObject::AutoStart(void) {
  // This is used to handle timers that start off enabled.
  // We do it this way because we don't want to actually start
  // the timer until we're ready.
  if(autostart)
    SetEnabled(true);
}

static Uint32 timer_callback(Uint32 interval, void *param) {
  ((TimerObject *)param)->Fire();
  return interval;
}


/* Property get/set functions */

// -- interval --

const char *TimerObject_GetInterval(ebCustomObject *sender) {
  return inttostr( ((TimerObject *)sender)->interval );
}

void TimerObject_SetInterval(ebCustomObject *sender, const char *value) {
  ((TimerObject *)sender)->SetInterval(atoi(value));
}

// -- enabled --

const char *TimerObject_GetEnabled(ebCustomObject *sender) {
  return booltostr( ((TimerObject *)sender)->enabled );
}

void TimerObject_SetEnabled(ebCustomObject *sender, const char *value) {
  ((TimerObject *)sender)->SetEnabled(strtobool(value));
}

/* Method functions */

const char *TimerObject_Start(ebCustomObject *sender, const char *method, int numargs, const char *args[]) {
  ((TimerObject *)sender)->SetEnabled(true);
  return "";
}

const char *TimerObject_Stop(ebCustomObject *sender, const char *method, int numargs, const char *args[]) {
  ((TimerObject *)sender)->SetEnabled(false);
  return "";
}
