/***************************************************************************
                          utils.cpp  -  eboxy general utility functions
                             -------------------
    begin                : Thu Aug 15 2002
    copyright            : (C) 2002 by Paul Eggleton
    email                : bluelightning@bluelightning.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <cstring>
#include <iostream>
#include <cstdlib>
#include <cstdio>

#include <ctype.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <limits.h>
#include <pwd.h>
#include <errno.h>

#include "SDL.h"
#include "utils.h"

using namespace std;

// Allocates space for a new string using malloc
char *newstring(int length) {
  // We use length+1 here to allow for the null terminator
  char *ret = (char *)malloc(length+1);
  if(ret) {
    ret[length] = '\0'; // for sanity's sake
    return ret;
  }
  else {
    // You'd hope this would never happen...
    cerr << "malloc("<< length+1 << ") failed" << endl;
    return NULL;
  }
}

// Trim off the whitespace at the start and end of the string
char *trimString(const char *string) {
  char *newStart;
  const char *p;
  int lastpos;

  // Find the first non-whitespace character
  p = string;
  for(int i=0; i < strlen(string); i++) {
    if(isspace(string[i]))
      p++;
    else
      break;
  }

  // Zero out all whitespace at the end of the string
  newStart = strdup(p);
  lastpos = strlen(newStart) - 1;
  for(int i = lastpos; i > 0; i--) {
    if(isspace(newStart[i]))
      newStart[i] = '\0';
    else
      break;
  }

  return newStart;
}

// Convert an HTML colour string (#rrggbb) into an SDL_Color structure
SDL_Color strtocolor(const char *colorstr) {
  SDL_Color color;
  int h;

  if(colorstr != NULL && strlen(colorstr) > 0 && colorstr[0] == '#') {
    sscanf(colorstr+1, "%X", &h);
    color.r = (h & 0xFF0000) >> 16;
    color.g = (h & 0xFF00) >> 8;
    color.b = (h & 0xFF);
  }
  else {
    color.r = 0;
    color.g = 0;
    color.b = 0;
  }
  return color;
}

// Convert an SDL_Color structure into an HTML colour string (#rrggbb)
char *colortostr(SDL_Color color) {
  char *colorstr = (char *)malloc(10);
  snprintf(colorstr, 8, "#%02X%02X%02X", color.r, color.g, color.b);
  return colorstr;
}

// Extracts the path component of a path+filename
char *getFilePath(const char *filename) {
  char *path = strdup(filename);
  char *p = strrchr(path, '/');
  if(p)
    *p = '\0';
  else
    strcpy(path, ".");
  return path;
}

// Expands any environment variables in the specified path
// Note: you should free what this returns when you're done with it
char *expandPath(const char *filename) {
  const int bufsize = PATH_MAX;

  if(filename == NULL)
    return NULL;

  char *path = strdup(filename);
  if(strlen(path) == 0)
    return path;

  char *buf = newstring(bufsize);
  char *bufptr;
  char *expandbuf = newstring(bufsize);
  bool expanding = false;
  int i=0, j=0, k=0;
  int expandstart = 0;

  while(true) {
    if(expanding) {
      if(i >= strlen(path) ||
          k >= bufsize ||
          !(isalnum(path[i]) || path[i] == '_')) {
        // End of variable, try to dereference it
        if(path[expandstart] == '$') {
          if(k==0 && i < strlen(path) && path[i] == '$') {
            // Special variable (process ID)
            bufptr = newstring(10);
            snprintf(bufptr, 10, "%d", getpid());
            int length = strlen(bufptr);
            if(bufsize-j < length)
              length = bufsize-j;
            strncat(buf, bufptr, length);
            j += length;
            free(bufptr);
            i++; // need to skip second $
          }
          else {
            // Get value of specified environment variable
            expandbuf[k] = '\0';
            bufptr = getenv(expandbuf);
            if(bufptr) {
              // getenv returned something, add it to the string
              int length = strlen(bufptr);
              if(bufsize-j < length)
                length = bufsize-j;
              buf[j] = '\0';
              strncat(buf, bufptr, length);
              j += length;
              // we DO NOT free bufptr here because it's a pointer into the user's env table
            }
          }
        }
        else if(path[expandstart] == '~') {
          // Note that we always expect j==0 here because anything else is not allowed
          if(i >= strlen(path) || path[i] == '/') {
            // Home directory
            if(k==0) {
              // Current user's home directory
              bufptr = getenv("HOME");
              if(bufptr) {
                strncpy(buf, bufptr, bufsize);
                j = strlen(bufptr);
                // we DO NOT free bufptr here because it's a pointer into the user's env table
              }
            }
            else {
              // Some other user's home directory
              expandbuf[k] = '\0';
              passwd *userinfo;
              userinfo = getpwnam(expandbuf);
              if(userinfo) {
                strncpy(buf, userinfo->pw_dir, bufsize);
                j = strlen(userinfo->pw_dir);
                //free(userinfo);   // why doesn't this work ?!
              }
              else {
                j = i;
                strncpy(buf, path, j);
              }
            }
          }
          else {
            j = i;
            strncpy(buf, path, j);
          }
        }
        expanding = false;
      }
      else
        expandbuf[k++] = path[i];
    }

    if(i >= strlen(path))
      break;

    if(!expanding) {
      if(path[i] == '$' || (i==0 && path[i] == '~')) {
        // Start of something to expand
        k = 0;
        expandstart = i;
        expanding = true;
      }
      else {
        buf[j] = path[i];
        j++;
      }
    }

    i++;
  }

  // Cleanup
  free(expandbuf);
  free(path);

  buf[j] = '\0';
  return buf;
}

// Convert a string value into a boolean
bool strtobool(const char *str) {
  if(strcasecmp(str, "true") == 0 || strcasecmp(str, "on") == 0 || strcasecmp(str, "yes") == 0 || atoi(str) != 0)
    return true;
  else
    return false;
}

// Convert boolean into a string value
const char *booltostr(bool value) {
  if(value)
    return "true";
  else
    return "false";
}

// Count the number of instances of character c in string str
int countstr(const char *str, char c) {
  int count = 0;
  const char *p = str;
  while(true) {
    p = strchr(p, c);
    if(p) {
      count++;
      p++;
    }
    else
      break;
  }
  return count;
}

// Convert an integer to a string
char *inttostr(int n) {
  char buf[16];
  snprintf(buf, 16, "%d", n);
  return strdup(buf);
}

// Find out if a point is inside a rectangle
bool pointinrect(int x, int y, SDL_Rect a) {
  return (x >= a.x && x <= (a.x + a.w) && y >= a.y && y <= (a.y + a.h));
}

// Find out if two SDL_Rects intersect
bool rectsintersect(SDL_Rect l, SDL_Rect r) {
  if(pointinrect(l.x, l.y, r) || pointinrect(l.x + l.w, l.y + l.h, r))
    return true;
  if(pointinrect(l.x, l.y + l.h, r) || pointinrect(l.x + l.w, l.y, r))
    return true;
  else if(l.x >= r.x && l.x <= (r.x + r.w) && r.y >= l.y && r.y <= (l.y + l.h))
    return true;
  else if(r.x >= l.x && r.x <= (l.x + l.w) && l.y >= r.y && l.y <= (r.y + r.h))
    return true;
  else
    return false;
}

// Join together two strings
char *strjoin(const char *str1, const char *str2) {
  char *result;

  if(str1!=NULL && str2!=NULL) {
    result = (char *)malloc(strlen(str1) + strlen(str2) + 1);
    strcpy(result, str1);
    strcat(result, str2);
    return result;
  }
  if(str1!=NULL && str2==NULL)
    return strdup(str1);
  if(str1==NULL && str2!=NULL)
    return strdup(str2);
  else
    return strdup("");
}

int pushUserEvent(int code, void *data1, void *data2) {
  // Push a user event into the SDL event queue (to be processed in the app's event loop)
  SDL_Event e;
  e.type = SDL_USEREVENT;
  e.user.type = SDL_USEREVENT;
  e.user.code = code;
  e.user.data1 = data1;
  e.user.data2 = data2;
  return SDL_PushEvent(&e);
}

int statfile(const char *filename) {
  struct stat buf;

  if(stat(filename, &buf) != 0)
    return errno;
  else
    return 0;
}

bool isEmptyString(const char *str) {
  for(int i=0;i<strlen(str);i++) {
    if(!isspace(str[i]))
      return false;
  }
  return true;
}

SDL_Color *colordup(const SDL_Color *c) {
  if(c) {
    SDL_Color *d = new SDL_Color;
    d->r = c->r;
    d->g = c->g;
    d->b = c->b;
    d->unused = c->unused;
    return d;
  }
  else
    return NULL;
}

template <class T> int vectorFind(vector<T> v, T obj) {
  for(int i=0; i<v.size(); i++) {
    if(v[i] == obj)
      return i;
  }
  return -1;
}
